from __future__ import annotations

import subprocess
from typing import List, Optional

_default_appname = "orcsome3"


def notify(summary: str, body: str, timeout: int = -1, urgency: int = 1, appname: Optional[str] = None) -> Notification:
    n: Notification = Notification(
        summary=summary, body=body, timeout=timeout, urgency=urgency, appname=appname or _default_appname
    )
    n.show()
    return n


class Notification(object):
    def __init__(self, summary: str, body: str, timeout: int, urgency: int, appname: str) -> None:
        self.summary: str = summary.lstrip("-")
        self.body: str = body
        self.timeout: int = timeout
        self.urgency: int = urgency
        self.appname: str = appname
        self.replace_id: int = 0
        self.lastcmd: List[str] = []

    def show(self) -> None:
        timeout: int = self.timeout * 1000
        if timeout < 0:
            timeout = -1

        urgency: str = "{}"
        if self.urgency != 1:
            urgency = f"{'urgency': <byte {self.urgency}>}"

        cmd: List[str] = [
            "gdbus",
            "call",
            "--session",
            "--dest=org.freedesktop.Notifications",
            "--object-path=/org/freedesktop/Notifications",
            "--method=org.freedesktop.Notifications.Notify",
            self.appname,
            str(self.replace_id),
            "",
            self.summary,
            self.body,
            "[]",
            urgency,
            str(timeout),
        ]
        self.lastcmd = cmd

        out, error = subprocess.Popen(args=cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).communicate()

        if error:
            raise Exception(error)

        self.replace_id = int(out.strip().split()[1].rstrip(b",)"))

    def update(
        self,
        summary: Optional[str] = None,
        body: Optional[str] = None,
        timeout: Optional[int] = None,
        urgency: Optional[int] = None,
    ) -> None:
        if summary is not None:
            self.summary = summary

        if body is not None:
            self.body = body

        if timeout is not None:
            self.timeout = timeout

        if urgency is not None:
            self.urgency = urgency

        self.show()

    def close(self) -> None:
        cmd: List[str] = [
            "gdbus",
            "call",
            "--session",
            "--dest=org.freedesktop.Notifications",
            "--object-path=/org/freedesktop/Notifications",
            "--method=org.freedesktop.Notifications.CloseNotification",
            str(self.replace_id),
        ]

        _, error = subprocess.Popen(args=cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT).communicate()
        if error:
            raise Exception(error)



from . import ev_build as ev_build
from ._ev import ffi as ffi, lib as lib
from typing import Any, Callable, Optional

class Loop:
    def __init__(self) -> None: ...
    def destroy(self) -> None: ...
    def run(self, flags: int = ...) -> None: ...
    def break_(self, flags: int = ...) -> None: ...

class IOWatcher:
    def __init__(self, callback: Callable[..., Any], file_descriptor: int, flags: int) -> None: ...
    def start(self, loop: Loop) -> None: ...
    def stop(self, loop: Loop) -> None: ...

class SignalWatcher:
    def __init__(self, callback: Callable[..., Any], signum: int) -> None: ...
    def start(self, loop: Loop) -> None: ...
    def stop(self, loop: Loop) -> None: ...

class TimerWatcher:
    next_stop: int
    def __init__(self, callback: Callable[..., Any], after: float, repeat: float = ...) -> None: ...
    def start(self, loop: Loop, after: Optional[float] = ..., repeat: Optional[float] = ...) -> None: ...
    def stop(self, loop: Loop) -> None: ...
    def again(self, loop: Loop) -> None: ...
    def remaining(self, loop: Loop) -> float: ...
    def update_next_stop(self) -> None: ...
    def overdue(self, timeout: float) -> bool: ...



import logging
import os
import signal
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Dict, Optional, Union

from ..version import VERSION
from . import ev, update_wm
from .wm import WM

logger: logging.Logger = logging.getLogger(name=__name__)


def execfile(filepath: Path, globales: Optional[Dict[str, Any]] = None) -> None:
    if globales is None:
        globales = globals()
    with filepath.open(mode="r") as fh:
        exec(fh.read() + "\n", globales)


def load_config(wm: WM, config: Path) -> None:
    update_wm(new_wm=wm)
    sys.path.insert(0, str(config.parent))
    try:
        execfile(filepath=config)
    except:
        logger.exception(msg=f"Error on loading {config}")
        sys.exit(1)
    finally:
        sys.path.pop(0)


"""def check_config(config: Path) -> bool:
    wm: TestWM = TestWM() # Mock class
    _update_wm(new_wm=wm)
    sys.path.insert(0, str(config.parent))
    try:
        execfile(filepath=config)
    except:
        logger.exception(msg=f"Config file check failed {config}")
        return False
    finally:
        sys.path.pop(0)

    return True"""


def run() -> None:
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument("--version", action="version", version="%(prog)s " + VERSION)
    parser.add_argument("-l", "--log", dest="log", metavar="FILE", help="Path to log file (log to stdout by default)")
    parser.add_argument("--log-level", metavar="LOGLEVEL", default="INFO", help="log level, default is INFO")

    config_dir: str = os.getenv(key="XDG_CONFIG_HOME", default=str(Path("~/.config").expanduser()))
    default_rcfile: str = str(Path(config_dir).joinpath("orcsome3", "rc.py"))
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        metavar="FILE",
        default=default_rcfile,
        help="Path to config file (%(default)s)",
    )

    args: Namespace = parser.parse_args()
    handler: Optional[Union[logging.FileHandler, logging.StreamHandler]] = None
    if args.log:
        handler = logging.FileHandler(filename=args.log)
    else:
        handler = logging.StreamHandler()

    root_logger: logging.Logger = logging.getLogger()
    root_logger.setLevel(level=args.log_level)
    handler.setFormatter(fmt=logging.Formatter(fmt="%(asctime)s %(name)s %(levelname)s: %(message)s"))
    root_logger.addHandler(hdlr=handler)

    if not Path(args.config).is_file():
        logger.info(msg="There is no config file available, exiting...")
        return

    loop: ev.Loop = ev.Loop()
    wm: WM = WM(loop=loop)

    def stop(loop_: Any, watcher: Any, events: int) -> None:
        wm.stop(is_exit=True)
        loop.break_()

    signal_watcher = ev.SignalWatcher(callback=stop, signum=signal.SIGINT)
    signal_watcher.start(loop=loop)

    def on_restart() -> None:
        wm.stop()
        logger.info(msg="Restarting...")
        load_config(wm=wm, config=Path(args.config))
        wm.init()
        logger.info(msg="Started successfully")

    wm._restart_handler = on_restart

    load_config(wm=wm, config=Path(args.config))
    wm.init()
    loop.run()



from typing import List, Optional


def notify(summary: str, body: str, timeout: int = ..., urgency: int = ..., appname: Optional[str] = ...) -> Notification: ...

class Notification:
    summary: str
    body: str
    timeout: int
    urgency: int
    appname: str
    replace_id: int
    lastcmd: List[str]
    def __init__(self, summary: str, body: str, timeout: int, urgency: int, appname: str) -> None: ...
    def show(self) -> None: ...
    def update(self, summary: Optional[str] = ..., body: Optional[str] = ..., timeout: Optional[int] = ..., urgency: Optional[int] = ...) -> None: ...
    def close(self) -> None: ...



from typing import Optional

from . import wm, wrappers


class Actions(wm.Actions):
    def __init__(self, window_manager: wm.WM) -> None:
        self._wm: wm.WM = window_manager

    def _focus(self, window: wrappers.Window, direction: int) -> None:
        clients = self._wm.find_clients(clients=self._wm.get_clients(), desktop=window.desktop)
        idx = clients.index(window)
        newc = clients[(idx + direction) % len(clients)]
        self._wm.focus_and_raise(window=newc)

    def focus_next(self, window: Optional[wrappers.Window] = None) -> None:
        window_: Optional[wrappers.Window] = window
        if window is None:
            window_ = self._wm.current_window
        if window_ is not None:
            self._focus(window=window_, direction=1)

    def focus_prev(self, window: Optional[wrappers.Window] = None) -> None:
        window_: Optional[wrappers.Window] = window
        if window is None:
            window_ = self._wm.current_window
        if window_ is not None:
            self._focus(window=window_, direction=-1)

    def restart(self) -> None:
        raise wm.RestartException()

    def activate_window_desktop(self, window: wrappers.Window) -> Optional[bool]:
        wd = window.desktop
        if wd is not None:
            if self._wm.current_desktop != wd:
                self._wm.activate_desktop(num=wd)
                return True
            else:
                return False
        else:
            return None



from __future__ import annotations

from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Tuple, Union, cast

from . import utils, wm, xlib


class WindowTree:
    def __init__(self, window: Window, root: Window, parent: Window, children: List[Window]) -> None:
        self.window: Window = window
        self.root: Window = root
        self.parent: Window = parent
        self.children: List[Window] = children


class XWindowAttributes:
    class MapState(Enum):
        IsUnmapped = int(xlib.lib.IsUnmapped)
        IsUnviewable = int(xlib.lib.IsUnviewable)
        IsViewable = int(xlib.lib.IsViewable)

    def __init__(self, attributes: xlib.XWindowAttributes) -> None:
        self.x: int = attributes.x  # location of window
        self.y: int = attributes.y  # location of window
        self.width: int = attributes.width  # width of window
        self.height: int = attributes.width  # height of window
        self.border_width: int = attributes.border_width  # border width of window
        self.depth: int = attributes.depth  # depth of window
        self.root: xlib.Window = attributes.root  # root of screen containing window
        self.override_redirect: bool = bool(attributes.override_redirect)  # boolean value for override-redirect
        self.map_state: XWindowAttributes.MapState = self.MapState(attributes.map_state)


class XScreenSaverInfo:
    class State(Enum):
        Off = int(xlib.lib.ScreenSaverOff)
        On = int(xlib.lib.ScreenSaverOn)
        Disabled = int(xlib.lib.ScreenSaverDisabled)

    class Kind(Enum):
        Blanked = int(xlib.lib.ScreenSaverBlanked)
        Internal = int(xlib.lib.ScreenSaverInternal)
        External = int(xlib.lib.ScreenSaverExternal)

    def __init__(self, screensaverinfo: xlib.ScreenSaverInfo) -> None:
        self.window: xlib.Window = screensaverinfo.window  # screen saver window
        self.state: XScreenSaverInfo.State = self.State(screensaverinfo.state)  # ScreenSaver{Off,On,Disabled}
        self.kind: XScreenSaverInfo.Kind = self.Kind(screensaverinfo.kind)  # ScreenSaver{Blanked,Internal,External}
        self.til_or_since: int = screensaverinfo.til_or_since  # time til or since screen saver (milliseconds)
        self.idle: int = screensaverinfo.idle  # total time since last user input (milliseconds)
        self.event_mask: int = screensaverinfo.eventMask  # currently selected events for this client


class Window(int):
    wm: wm.WM

    @cached_property
    def desktop(self) -> Optional[int]:
        """Return window desktop.

        Result is:

        * number from 0 to desktop_count - 1
        * -1 if window placed on all desktops
        * None if window does not have desktop property

        """
        result = self.get_property(property="_NET_WM_DESKTOP", type="CARDINAL")
        if not result:
            return None
        desktop = cast(List[int], result)[0]
        if desktop == 0xFFFFFFF or desktop == 0xFFFFFFFFFFFFFFFF:
            desktop = -1
        return desktop

    @cached_property
    def role(self) -> Optional[str]:
        """Return WM_WINDOW_ROLE property"""
        result = self.get_property(property="WM_WINDOW_ROLE", type="STRING")
        if not result or not len(result):
            return None
        return str(result[0])

    @cached_property
    def cls(self) -> Optional[str]:
        """Return second part from WM_CLASS property"""
        name, _class = self.get_name_and_class()
        if name:
            setattr(self, "name", name)
        return _class if _class else None

    @cached_property
    def name(self) -> Optional[str]:
        """Return first part from WM_CLASS property"""
        name, _class = self.get_name_and_class()
        if _class:
            setattr(self, "cls", _class)
        return name if name else None

    @cached_property
    def title(self) -> Optional[str]:
        """Return _NET_WM_NAME property"""
        result = self.get_property(property="_NET_WM_NAME", type="UTF8_STRING")
        if not result:
            return None
        return cast(List[str], result)[0]

    def get_name_and_class(self) -> Tuple[Optional[str], Optional[str]]:
        """Return WM_CLASS property"""
        result = self.get_property(property="WM_CLASS", type="STRING", split=True)
        if not result or len(result) != 2:
            return None, None
        result = cast(List[str], result)
        return result[0], result[1]

    def matches(
        self,
        name: Optional[str] = None,
        cls: Optional[str] = None,
        role: Optional[str] = None,
        desktop: Optional[int] = None,
        title: Optional[str] = None,
    ) -> bool:
        """Check if window suits given matchers.

        Matchers keyword arguments are used in :meth:`on_create`,
        :func:`actions.spawn_or_raise`. :meth:`find_clients` and
        :meth:`find_client`.

        name
          window name (also referenced as `instance`).
          The first part of ``WM_CLASS`` property.

        cls
          window class. The second part of ``WM_CLASS`` property.

        role
          window role. Value of ``WM_WINDOW_ROLE`` property.

        desktop
          matches windows placed on specific desktop. Must be int.

        title
          window title.

        `name`, `cls`, `title` and `role` can be regular expressions.

        """
        if name and not utils.match_string(pattern=name, data=self.name or ""):
            return False
        if cls and not utils.match_string(pattern=cls, data=self.cls or ""):
            return False
        if role and not utils.match_string(pattern=role, data=self.role or ""):
            return False
        if title and not utils.match_string(pattern=title, data=self.title or ""):
            return False
        if desktop is not None and desktop != self.desktop:
            return False

        return True

    def get_property(
        self, property: str, type: Optional[str] = None, split: bool = False
    ) -> Optional[Union[List[int], List[str]]]:
        """
        This function is a wrapper for `XGetWindowProperty`. Returns a property for the window, the result can be `None` if no property was found, a `List[int]`or a `List[str]`.
        """
        return xlib.get_window_property(
            display=self.wm.dpy,
            window=self,
            property=self.wm.atom[property],
            type=self.wm.atom[type] if type else xlib.lib.AnyPropertyType,
            split=split,
        )

    def set_property(
        self, property: str, format: int, data: Union[List[int], List[str]], type: Optional[str] = None
    ) -> None:
        """
        This functions is a wrapper for `XChangeProperty`, it alters the property for the specified window.

        If the specified format is 8, the property data must be a char array. (List[str] on python)

        If the specified format is 16, the property data must be a short array. (List[int] on python)

        If the specified format is 32, the property data must be a long array. (List[int] on python)
        """
        xlib.set_window_property(
            display=self.wm.dpy,
            window=self,
            property=self.wm.atom[property],
            type=self.wm.atom[type] if type else xlib.lib.AnyPropertyType,
            format=format,
            values=data,
        )

    def get_windows_same_pid(self) -> List[Window]:
        """
        This function returns a List[Window] that has the same pid of `self`
        """
        windows_associated: List[Window] = []
        window_tree: Optional[Tuple[Window, Window, List[Window], int]] = self.wm._get_window_tree(
            window=self.wm.create_window(window_id=self.wm.root)
        )
        if not window_tree or not len(window_tree[2]) or not self.pid:
            return windows_associated
        for window in window_tree[2]:
            if window.pid == self.pid:
                windows_associated.append(window)
        return windows_associated

    def get_window_tree(self) -> Optional[WindowTree]:
        """
        This function is a wrapper for `XQueryTree`. Returns a WindowTree instance containing
        the window, the root window, the parent window and a list of children windows (empty if there's none)
        """
        window_tree: Optional[Tuple[Window, Window, List[Window], int]] = self.wm._get_window_tree(window=self)
        if not window_tree:
            return None
        return WindowTree(window=self, root=window_tree[0], parent=window_tree[1], children=window_tree[2])

    def set_window_icon(self, icon: Union[Path, str]) -> None:
        """
        This function sets the window's icon, the maximum icon size is 10Mb
        """
        if isinstance(icon, str):
            icon = Path(icon)
        if not icon.is_file():
            print("The path is not a valid file")
            return
        if icon.stat().st_size > 10000000:
            print("The maximum icon size is 10Mb")
            return
        self.wm._set_window_icon(window=self, icon=str(icon))

    @cached_property
    def attributes(self) -> XWindowAttributes:
        """
        This function is a wrapper for `XGetWindowAttributes`. Returns the current attributes for the window
        """
        attrs = xlib.get_window_attributes(display=self.wm.dpy, window=self)
        return XWindowAttributes(attributes=attrs)

    @cached_property
    def state(self) -> List[str]:
        """Return _NET_WM_STATE"""
        states = self.get_property(property="_NET_WM_STATE", type="ATOM")
        if not states:
            return []
        states = cast(List[xlib.Atom], states)
        return [self.wm.get_atom_name(atom=state) for state in states if self.wm.get_atom_name(atom=state)]

    @cached_property
    def maximized_vert(self) -> bool:
        return "_NET_WM_STATE_MAXIMIZED_VERT" in self.state

    @cached_property
    def maximized_horz(self) -> bool:
        return "_NET_WM_STATE_MAXIMIZED_HORZ" in self.state

    @cached_property
    def decorated(self) -> bool:
        """
        Returns False if the window is not decorated otherwise True

        A non-decorated window is considered when:
        - There's no window manager running
        - It has the attribute override-redirect
        - It has a 0 in the third bit on the property `_MOTIF_WM_HINTS`
        """
        decorated: bool = True
        try:
            if not self.wm.wm_name:  # If the window manager is not running window can't be decorated
                decorated = False
            else:
                # If the window has the attribute override-redirect it can't be decorated
                if self.attributes.override_redirect:
                    decorated = False
                # If the window manager running is Openbox then it checks if window has the state '_OB_WM_STATE_UNDECORATED'
                elif self.wm.wm_name == "Openbox":
                    if "_OB_WM_STATE_UNDECORATED" in self.state:
                        decorated = False
                # If the window manager is not Openbox then it looks for the third bit on the property '_MOTIF_WM_HINTS'
                # to determine the decorations
                else:
                    """
                    {
                        int     flags;
                        int     functions;
                        int     decorations;
                        int     input_mode;
                        int     status;
                    } MotifWmHints;
                    """
                    motif_hints = self.get_property(property="_MOTIF_WM_HINTS", type="_MOTIF_WM_HINTS")
                    # If the decorations bit in the _MOTIF_WM_HINTS (position 2) is 0 then we can assume
                    # the window is not decorated
                    if motif_hints is not None and cast(List[int], motif_hints)[2] == 0:
                        decorated = False
        except:
            decorated = False
        return decorated

    @cached_property
    def urgent(self) -> bool:
        return "_NET_WM_STATE_DEMANDS_ATTENTION" in self.state

    @cached_property
    def fullscreen(self) -> bool:
        return "_NET_WM_STATE_FULLSCREEN" in self.state

    @cached_property
    def pid(self) -> Optional[int]:
        result = self.get_property(property="_NET_WM_PID", type="CARDINAL")
        if not result:
            return None
        return cast(List[int], result)[0]



from typing import Optional, cast

from .wm import WM, ImmediateWM

_wm: Optional[WM] = None


def get_wm() -> WM:
    return cast(WM, _wm)


def get_wm_immediate() -> ImmediateWM:
    return ImmediateWM()


def update_wm(new_wm: WM) -> None:
    global _wm
    _wm = new_wm



from . import utils as utils, wm as wm, xlib as xlib
from enum import Enum
from functools import cached_property as cached_property
from pathlib import Path
from typing import List, Optional, Tuple, Union

class WindowTree:
    window: Window
    root: Window
    parent: Window
    children: List[Window]
    def __init__(self, window: Window, root: Window, parent: Window, children: List[Window]) -> None: ...

class XWindowAttributes:
    class MapState(Enum):
        IsUnmapped: Any
        IsUnviewable: Any
        IsViewable: Any
    x: int
    y: int
    width: int
    height: int
    border_width: int
    depth: int
    root: Window
    override_redirect: bool
    map_state: XWindowAttributes.MapState
    def __init__(self, attributes: xlib.XWindowAttributes) -> None: ...

class XScreenSaverInfo:
    class State(Enum):
        Off: Any
        On: Any
        Disabled: Any
    class Kind(Enum):
        Blanked: Any
        Internal: Any
        External: Any
    window: Window
    state: XScreenSaverInfo.State
    kind: XScreenSaverInfo.Kind
    til_or_since: int
    idle: int
    event_mask: int
    def __init__(self, screensaverinfo: xlib.ScreenSaverInfo) -> None: ...

class Window(int):
    wm: wm.WM
    @cached_property
    def desktop(self) -> Optional[int]: ...
    @cached_property
    def role(self) -> Optional[str]: ...
    @cached_property
    def cls(self) -> Optional[str]: ...
    @cached_property
    def name(self) -> Optional[str]: ...
    @cached_property
    def title(self) -> Optional[str]: ...
    def get_name_and_class(self) -> Tuple[Optional[str], Optional[str]]: ...
    def matches(self, name: Optional[str] = ..., cls: Optional[str] = ..., role: Optional[str] = ..., desktop: Optional[int] = ..., title: Optional[str] = ...) -> bool: ...
    def get_property(self, property: str, type: Optional[str] = ..., split: bool = ...) -> Optional[Union[List[int], List[str]]]: ...
    def set_property(self, property: str, format: int, data: Union[List[int], List[str]], type: Optional[str] = ...) -> None: ...
    def get_windows_same_pid(self) -> List[Window]: ...
    def get_window_tree(self) -> Optional[WindowTree]: ...
    def set_window_icon(self, icon: Union[Path, str]) -> None: ...
    @cached_property
    def attributes(self) -> XWindowAttributes: ...
    @cached_property
    def state(self) -> List[str]: ...
    @cached_property
    def maximized_vert(self) -> bool: ...
    @cached_property
    def maximized_horz(self) -> bool: ...
    @cached_property
    def decorated(self) -> bool: ...
    @cached_property
    def urgent(self) -> bool: ...
    @cached_property
    def fullscreen(self) -> bool: ...
    @cached_property
    def pid(self) -> Optional[int]: ...



import logging
from . import ev as ev, update_wm as update_wm
from ..version import VERSION as VERSION
from .wm import WM as WM
from pathlib import Path
from typing import Any, Dict, Optional

logger: logging.Logger

def execfile(filepath: Path, globales: Optional[Dict[str, Any]] = ...) -> None: ...
def load_config(wm: WM, config: Path) -> None: ...
def run() -> None: ...



from . import xlib_build as xlib_build
from ._xlib import ffi as ffi, lib as lib
from enum import Enum
from typing import Any, Dict, List, Optional, Union

Atom = int
Window = int
Display = Any
XEvent = Any
XErrorEvent = Any
XWindowAttributes = Any
ScreenSaverInfo = Any

class MASKS(Enum):
    Mod1Mask: Any
    ControlMask: Any
    ShiftMask: Any
    Mod2Mask: Any
    Mod4Mask: Any
    LockMask: Any

class EVENTS(Enum):
    KeyPress: Any
    KeyRelease: Any
    CreateNotify: Any
    DestroyNotify: Any
    FocusIn: Any
    FocusOut: Any
    PropertyNotify: Any
    MapNotify: Any
    UnmapNotify: Any

class AtomCache:
    dpy: Display
    def __init__(self, dpy: Display) -> None: ...
    def __getitem__(self, name: str) -> Atom: ...

class XEvent_:
    class Type(Enum):
        KeyPress: Any
        KeyRelease: Any
        CreateNotify: Any
        DestroyNotify: Any
        PropertyNotify: Any
        FocusIn: Any
        FocusOut: Any
        CirculateNotify: Any
        ConfigureNotify: Any
        GravityNotify: Any
        ReparentNotify: Any
        MapNotify: Any
        UnmapNotify: Any
    type: XEvent_.Type
    def __init__(self, xevent: XEvent, specific_event: Any) -> None: ...

class XKeyEvent(XEvent_):
    window: Window
    root: Window
    subwindow: Window
    time: float
    x: int
    y: int
    x_root: int
    y_root: int
    state: int
    keycode: int
    same_screen: bool
    def __init__(self, event: XEvent) -> None: ...

class XCreateWindowEvent(XEvent_):
    parent: Window
    window: Window
    x: int
    y: int
    width: int
    height: int
    border_width: int
    override_redirect: bool
    def __init__(self, event: XEvent) -> None: ...

class XDestroyWindowEvent(XEvent_):
    parent: Window
    window: Window
    def __init__(self, event: XEvent) -> None: ...

class XPropertyEvent(XEvent_):
    class State(Enum):
        PropertyNewValue: Any
        PropertyDelete: Any
    window: Window
    atom: Atom
    time: float
    state: int
    def __init__(self, event: XEvent) -> None: ...

class XFocusChangeEvent(XEvent_):
    class Mode(Enum):
        NotifyNormal: Any
        NotifyWhileGrabbed: Any
        NotifyGrab: Any
        NotifyUngrab: Any
    class Detail(Enum):
        NotifyAncestor: Any
        NotifyVirtual: Any
        NotifyInferior: Any
        NotifyNonlinear: Any
        NotifyNonlinearVirtual: Any
        NotifyPointer: Any
        NotifyPointerRoot: Any
        NotifyDetailNone: Any
    window: Window
    mode: XFocusChangeEvent.Mode
    detail: XFocusChangeEvent.Detail
    def __init__(self, event: XEvent) -> None: ...

class XErrorEvent_:
    type: XEvent_.Type
    display: Display
    resourceid: float
    serial: float
    error_code: str
    request_code: str
    minor_code: str
    def __init__(self, error: XErrorEvent) -> None: ...
    def get_message(self, size: int = ...) -> str: ...

def get_window_property(display: Display, window: Window, property: Atom, type: Atom = ..., split: bool = ...) -> Optional[Union[List[int], List[str]]]: ...
def get_window_attributes(display: Display, window: Window) -> Optional[XWindowAttributes]: ...
def get_screen_saver_info(display: Display, drawable: Window) -> Optional[ScreenSaverInfo]: ...
def set_window_property(display: Display, window: Window, property: Atom, type: Atom, format: int, values: Union[List[int], List[str]]) -> None: ...
def get_kbd_group(display: Display) -> str: ...
def set_kbd_group(display: Display, group: int) -> None: ...
def get_atom_name(display: Display, atom: Atom) -> str: ...



import re
import subprocess
from typing import Dict, List, Pattern, Tuple

_re_cache: Dict[str, Pattern[str]] = {}


def match_string(pattern: str, data: str) -> bool:
    if not pattern in _re_cache.keys():
        _re_cache[pattern] = re.compile(pattern=pattern)
    pattern_ = _re_cache[pattern]

    return bool(re.search(pattern=pattern_, string=data))


def get_compiler_args(*libraries: str) -> Tuple[List[str], List[str]]:
    cmd_cflags: List[str] = ["pkg-config", "--cflags"]
    cmd_cflags.extend(libraries)
    process_cflags: subprocess.CompletedProcess = subprocess.run(
        args=cmd_cflags, capture_output=True, text=True, encoding="utf-8"
    )
    result_cflags: List[str] = [x for x in process_cflags.stdout.replace("\n", "").split(sep=" ") if len(x.strip())]

    cmd_libs: List[str] = ["pkg-config", "--libs"]
    cmd_libs.extend(libraries)
    process_libs: subprocess.CompletedProcess = subprocess.run(
        args=cmd_libs, capture_output=True, text=True, encoding="utf-8"
    )
    result_libs: List[str] = [
        x.replace("-l", "") for x in process_libs.stdout.replace("\n", "").split(sep=" ") if len(x.strip())
    ]

    return result_cflags, result_libs



import abc
import logging
from . import ev as ev, wrappers as wrappers, xlib as xlib
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

logger: logging.Logger
ignore_logger: bool
MODIFICATORS: Dict[str, int]
IGNORED_MOD_MASKS: Tuple[int, int, int, int]

class RestartException(Exception): ...

class Actions(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def focus_next(self, window: Optional[wrappers.Window] = ...) -> None: ...
    @abstractmethod
    def focus_prev(self, window: Optional[wrappers.Window] = ...) -> None: ...
    @abstractmethod
    def restart(self) -> None: ...
    @abstractmethod
    def activate_window_desktop(self, window: wrappers.Window) -> Optional[bool]: ...

class WM:
    focus_history: List[xlib.Window]
    dpy: xlib.Display
    root: xlib.Window
    atom: xlib.AtomCache
    track_kbd_layout: bool
    actions: Actions
    def __init__(self, loop: ev.Loop) -> None: ...
    def init(self) -> None: ...
    def stop(self, is_exit: bool = ...) -> None: ...
    def create_window(self, window_id: int) -> wrappers.Window: ...
    def get_keycode_from_string(self, key: str) -> Optional[int]: ...
    def parse_keydef(self, keydef: str) -> Optional[List[Tuple[int, int]]]: ...
    def on_key(self, keydef: str, window: Optional[wrappers.Window] = ...) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def on_create(self, **matchers: Any) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def on_manage(self, **matchers: Any) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def on_destroy(self, window: Optional[wrappers.Window] = ...) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def on_property_change(self, properties: List[str], window: Optional[wrappers.Window] = ...) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def on_timer(self, timeout: float, start: bool = ..., first_timeout: Optional[float] = ...) -> Callable[[Callable[[], None]], Callable[[], None]]: ...
    def get_clients(self) -> List[wrappers.Window]: ...
    def get_stacked_clients(self) -> List[wrappers.Window]: ...
    @property
    def current_window(self) -> Optional[wrappers.Window]: ...
    @property
    def current_desktop(self) -> int: ...
    @property
    def wm_name(self) -> Optional[str]: ...
    @property
    def event_window(self) -> wrappers.Window: ...
    @property
    def event(self) -> Union[xlib.XKeyEvent, xlib.XCreateWindowEvent, xlib.XDestroyWindowEvent, xlib.XPropertyEvent]: ...
    def activate_desktop(self, num: int) -> None: ...
    def find_clients(self, clients: List[wrappers.Window], **matchers: Any) -> List[wrappers.Window]: ...
    def find_client(self, clients: List[wrappers.Window], **matchers: Any) -> Optional[wrappers.Window]: ...
    def focus_window(self, window: xlib.Window) -> None: ...
    def focus_and_raise(self, window: wrappers.Window) -> None: ...
    def place_window_above(self, window: xlib.Window) -> None: ...
    def place_window_below(self, window: xlib.Window) -> None: ...
    def minimize_window(self, window: xlib.Window) -> None: ...
    def restore_window(self, window: xlib.Window) -> None: ...
    def set_window_state(self, window: xlib.Window, taskbar: Optional[bool] = ..., pager: Optional[bool] = ..., decorate: Optional[bool] = ..., otaskbar: Optional[bool] = ..., vmax: Optional[bool] = ..., hmax: Optional[bool] = ...) -> None: ...
    def get_window_geometry(self, window: xlib.Window) -> Tuple[int, int, int, int]: ...
    def get_screen_size(self) -> Tuple[int, int]: ...
    def get_workarea(self, desktop: Optional[int] = ...) -> List[int]: ...
    def moveresize_window(self, window: xlib.Window, x: Optional[int] = ..., y: Optional[int] = ..., w: Optional[int] = ..., h: Optional[int] = ...) -> None: ...
    def moveresize_window2(self, window: wrappers.Window, left: int, top: int, right: int, bottom: int) -> None: ...
    def close_window(self, window: Optional[wrappers.Window] = ...) -> None: ...
    def change_window_desktop(self, window: xlib.Window, desktop: int) -> None: ...
    def on_init(self, func: Callable[[], None]) -> Callable[[], None]: ...
    def on_deinit(self, func: Callable[[], None]) -> Callable[[], None]: ...
    def get_screen_saver_info(self) -> Optional[wrappers.XScreenSaverInfo]: ...
    def reset_dpms(self) -> None: ...
    def get_atom_name(self, atom: xlib.Atom) -> str: ...

class ImmediateWM(WM):
    dpy: xlib.Display
    root: xlib.Window
    atom: xlib.AtomCache
    def __init__(self) -> None: ...

def error_handler(_display: xlib.Display, error: xlib.XErrorEvent) -> int: ...



from typing import List, Tuple

import cffi
from cffi.api import FFI

from orcsome3.orcsome import utils

source: str = """
#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/scrnsaver.h>
#include <X11/extensions/dpms.h>
#include <X11/extensions/XKB.h>
#include <X11/extensions/XKBstr.h>
#include <MagickWand/MagickWand.h>

#define gdMaxColors 256
#define gdImageSX(im) ((im)->sx)
#define gdImageSY(im) ((im)->sy)
#define gdTrueColorGetBlue(c) ((c) & 0x0000FF)
#define gdTrueColorGetGreen(c) (((c) & 0x00FF00) >> 8)
#define gdTrueColorGetRed(c) (((c) & 0xFF0000) >> 16)
#define gdTrueColorGetAlpha(c) (((c) & 0x7F000000) >> 24)
#define gdImageBlue(im, c) ((im)->trueColor ? gdTrueColorGetBlue(c) : (im)->blue[(c)])
#define gdImageGreen(im, c) ((im)->trueColor ? gdTrueColorGetGreen(c) : (im)->green[(c)])
#define gdImageRed(im, c) ((im)->trueColor ? gdTrueColorGetRed(c) : (im)->red[(c)])
#define gdImageAlpha(im, c) ((im)->trueColor ? gdTrueColorGetAlpha(c) : (im)->alpha[(c)])

typedef enum {
    GD_DEFAULT = 0,
    GD_BELL,
    GD_BESSEL,
    GD_BILINEAR_FIXED,
    GD_BICUBIC,
    GD_BICUBIC_FIXED,
    GD_BLACKMAN,
    GD_BOX,
    GD_BSPLINE,
    GD_CATMULLROM,
    GD_GAUSSIAN,
    GD_GENERALIZED_CUBIC,
    GD_HERMITE,
    GD_HAMMING,
    GD_HANNING,
    GD_MITCHELL,
    GD_NEAREST_NEIGHBOUR,
    GD_POWER,
    GD_QUADRATIC,
    GD_SINC,
    GD_TRIANGLE,
    GD_WEIGHTED4,
    GD_LINEAR,
    GD_LANCZOS3,
    GD_LANCZOS8,
    GD_BLACKMAN_BESSEL,
    GD_BLACKMAN_SINC,
    GD_QUADRATIC_BSPLINE,
    GD_CUBIC_SPLINE,
    GD_COSINE,
    GD_WELSH,
    GD_METHOD_COUNT = 30
    } gdInterpolationMethod;

typedef double (* interpolation_method )(double, double);

typedef struct gdImageStruct {
    /* Palette-based image pixels */
    unsigned char **pixels;
    int sx;
    int sy;
    /* These are valid in palette images only. See also
        'alpha', which appears later in the structure to
        preserve binary backwards compatibility */
    int colorsTotal;
    int red[gdMaxColors];
    int green[gdMaxColors];
    int blue[gdMaxColors];
    int open[gdMaxColors];
    /* For backwards compatibility, this is set to the
        first palette entry with 100% transparency,
        and is also set and reset by the
        gdImageColorTransparent function. Newer
        applications can allocate palette entries
        with any desired level of transparency; however,
        bear in mind that many viewers, notably
        many web browsers, fail to implement
        full alpha channel for PNG and provide
        support for full opacity or transparency only. */
    int transparent;
    int *polyInts;
    int polyAllocated;
    struct gdImageStruct *brush;
    struct gdImageStruct *tile;
    int brushColorMap[gdMaxColors];
    int tileColorMap[gdMaxColors];
    int styleLength;
    int stylePos;
    int *style;
    int interlace;
    /* New in 2.0: thickness of line. Initialized to 1. */
    int thick;
    /* New in 2.0: alpha channel for palettes. Note that only
        Macintosh Internet Explorer and (possibly) Netscape 6
        really support multiple levels of transparency in
        palettes, to my knowledge, as of 2/15/01. Most
        common browsers will display 100% opaque and
        100% transparent correctly, and do something
        unpredictable and/or undesirable for levels
        in between. TBB */
    int alpha[gdMaxColors];
    /* Truecolor flag and pixels. New 2.0 fields appear here at the
        end to minimize breakage of existing object code. */
    int trueColor;
    int **tpixels;
    /* Should alpha channel be copied, or applied, each time a
        pixel is drawn? This applies to truecolor images only.
        No attempt is made to alpha-blend in palette images,
        even if semitransparent palette entries exist.
        To do that, build your image as a truecolor image,
        then quantize down to 8 bits. */
    int alphaBlendingFlag;
    /* Should the alpha channel of the image be saved? This affects
        PNG at the moment; other future formats may also
        have that capability. JPEG doesn't. */
    int saveAlphaFlag;

    /* There should NEVER BE ACCESSOR MACROS FOR ITEMS BELOW HERE, so this
        part of the structure can be safely changed in new releases. */

    /* 2.0.12: anti-aliased globals. 2.0.26: just a few vestiges after
        switching to the fast, memory-cheap implementation from PHP-gd. */
    int AA;
    int AA_color;
    int AA_dont_blend;

    /* 2.0.12: simple clipping rectangle. These values
        must be checked for safety when set; please use
        gdImageSetClip */
    int cx1;
    int cy1;
    int cx2;
    int cy2;

    /* 2.1.0: allows to specify resolution in dpi */
    unsigned int res_x;
    unsigned int res_y;

    /* Selects quantization method, see gdImageTrueColorToPaletteSetMethod() and gdPaletteQuantizationMethod enum. */
    int paletteQuantizationMethod;
    /* speed/quality trade-off. 1 = best quality, 10 = best speed. 0 = method-specific default.
        Applicable to GD_QUANT_LIQ and GD_QUANT_NEUQUANT. */
    int paletteQuantizationSpeed;
    /* Image will remain true-color if conversion to palette cannot achieve given quality.
        Value from 1 to 100, 1 = ugly, 100 = perfect. Applicable to GD_QUANT_LIQ.*/
    int paletteQuantizationMinQuality;
    /* Image will use minimum number of palette colors needed to achieve given quality. Must be higher than paletteQuantizationMinQuality
        Value from 1 to 100, 1 = ugly, 100 = perfect. Applicable to GD_QUANT_LIQ.*/
    int paletteQuantizationMaxQuality;
    gdInterpolationMethod interpolation_id;
    interpolation_method interpolation;
} gdImage;

typedef gdImage *gdImagePtr;

/* We can't use the one defined in Xmd.h because that's an "unsigned int",
 * which comes out as a 32bit type always. We need this to be 64bit on 64bit
 * machines.
 */
typedef unsigned long int CARD32_;

gdImagePtr gdImageCreateFromPngPtr (int size, void* data);
void gdImageDestroy(gdImagePtr im);
int gdImageGetPixel(gdImagePtr im, int x, int y);

void set_window_icon(Display *display, Window window, char *filepath) {
    gdImagePtr imagePtr = NULL; // Default value, if the image can't be loaded then the result is going to be NULL
    MagickWand *wand = NewMagickWand();
    PixelWand *pixelWand = NewPixelWand();
    
    PixelSetColor(pixelWand, "transparent"); // Color for the background
    // The background has to be assigned before loading the file image
    // to avoid white backgrounds (happens with svg files)
    MagickSetBackgroundColor(wand, pixelWand);
    
    int result = 0;
    result = MagickReadImage(wand, filepath); // Reads the file
    if (!result) {
        return;
    }
    result = MagickSetImageFormat(wand, "png"); // Converts the wand (image) to PNG format
    if (!result) {
        return;
    }
    // If the conversion is successful then it creates an imagePtr from the content of the wand
    size_t length;
    unsigned char *info = MagickGetImageBlob(wand, &length); // Gets image data and size from the wand
    imagePtr = gdImageCreateFromPngPtr((int) length, info); // Creates imagePtr from png data
    if (imagePtr == NULL) {
        return;
    }
    
    DestroyMagickWand(wand);
    DestroyPixelWand(pixelWand);

    int width, height;
    width = gdImageSX(imagePtr);
    height = gdImageSY(imagePtr);

    unsigned int ndata = (width * height) + 2;
    CARD32_* data = (CARD32_*) calloc(ndata, sizeof(CARD32_));

    int i = 0;
    data[i++] = width;
    data[i++] = height;

    int x, y;
    for(y = 0; y < height; y++) {
        for(x = 0; x < width; x++) {
            // data is RGBA
            // We'll do some horrible data-munging here
            unsigned char* cols = (unsigned char*) &(data[i++]);
            int pixcolour = gdImageGetPixel(imagePtr, x, y);

            cols[0] = gdImageBlue(imagePtr, pixcolour);
            cols[1] = gdImageGreen(imagePtr, pixcolour);
            cols[2] = gdImageRed(imagePtr, pixcolour);

            /* Alpha is more difficult */
            int alpha = 127 - gdImageAlpha(imagePtr, pixcolour); // 0 to 127
            // Scale it up to 0 to 255; remembering that 2*127 should be max
            cols[3] = alpha == 127 ? 255 : (alpha *= 2);
        }
    }
    gdImageDestroy(imagePtr);

    Atom property = XInternAtom(display, "_NET_WM_ICON", 0);
    Atom type = XInternAtom(display, "CARDINAL", 0);
    result = XChangeProperty(display, window, property, type, 32, PropModeReplace, (unsigned char*) data, (int) ndata);
    if (result) {
        XFlush(display);
    }
    XFree(data);
}
"""

export_source: str = """
static const long StructureNotifyMask;
static const long SubstructureNotifyMask;
static const long SubstructureRedirectMask;
static const long PropertyChangeMask;
static const long FocusChangeMask;
static const long KeyPressMask;
static const long KeyReleaseMask;

static const long CurrentTime;

static const int KeyPress;
static const int KeyRelease;
static const int CreateNotify;
static const int DestroyNotify;
static const int FocusIn;
static const int FocusOut;
static const int PropertyNotify;
static const int ClientMessage;

static const int CWX;
static const int CWY;
static const int CWWidth;
static const int CWHeight;
static const int CWBorderWidth;
static const int CWSibling;
static const int CWStackMode;

static const int Above;
static const int Below;

static const int ShiftMask;
static const int LockMask;
static const int ControlMask;
static const int Mod1Mask;
static const int Mod2Mask;
static const int Mod3Mask;
static const int Mod4Mask;
static const int Mod5Mask;
static const int AnyKey;
static const int AnyModifier;

static const long NoSymbol;

static const int GrabModeSync;
static const int GrabModeAsync;

static const int PropModeReplace;
static const int PropModePrepend;
static const int PropModeAppend;

static const int XkbUseCoreKbd;

static const int ScreenSaverOff;
static const int ScreenSaverOn;
static const int ScreenSaverDisabled;
static const int ScreenSaverBlanked;
static const int ScreenSaverInternal;
static const int ScreenSaverExternal;

static const int IsUnmapped;
static const int IsUnviewable;
static const int IsViewable;

static const int PropertyNewValue;
static const int PropertyDelete;

static const int NotifyNormal;
static const int NotifyWhileGrabbed;
static const int NotifyGrab;
static const int NotifyUngrab;

static const int NotifyAncestor;
static const int NotifyVirtual;
static const int NotifyInferior;
static const int NotifyNonlinear;
static const int NotifyNonlinearVirtual;
static const int NotifyPointer;
static const int NotifyPointerRoot;
static const int NotifyDetailNone;

static const int AnyPropertyType;

// Events related with SubstructureRedirectMask
static const int CirculateNotify;
static const int ConfigureNotify;
static const int GravityNotify;
static const int MapNotify;
static const int ReparentNotify;
static const int UnmapNotify;

typedef int Bool;
typedef int Status;
typedef unsigned long XID;
typedef unsigned long Time;
typedef unsigned long Atom;
typedef XID Window;
typedef XID Drawable;
typedef XID KeySym;
typedef XID Cursor;
typedef unsigned char KeyCode;
typedef ... Display;
typedef ... Visual;
typedef XID Colormap;
typedef ... Screen;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window window;          /* window on which event was requested in event mask */
} XAnyEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window window;
    Atom message_type;
    int format;
    union {
        char b[20];
        short s[10];
        long l[5];
    } data;
} XClientMessageEvent;

typedef struct {
    int type;               /* of event */
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window window;          /* "event" window it is reported relative to */
    Window root;            /* root window that the event occurred on */
    Window subwindow;       /* child window */
    Time time;              /* milliseconds */
    int x, y;               /* pointer x, y coordinates in event window */
    int x_root, y_root;     /* coordinates relative to root */
    unsigned int state;     /* key or button mask */
    unsigned int keycode;   /* detail */
    Bool same_screen;       /* same screen flag */
} XKeyEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window parent;          /* parent of the window */
    Window window;          /* window id of window created */
    int x, y;               /* window location */
    int width, height;      /* size of window */
    int border_width;       /* border width */
    Bool override_redirect; /* creation should be overridden */
} XCreateWindowEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window event;
    Window window;
} XDestroyWindowEvent;

typedef struct {
    int type;               /* FocusIn or FocusOut */
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window window;          /* window of event */
    int mode;               /* NotifyNormal, NotifyWhileGrabbed, NotifyGrab, NotifyUngrab */
    int detail;             /* NotifyAncestor, NotifyVirtual, NotifyInferior,
                            * NotifyNonlinear,NotifyNonlinearVirtual, NotifyPointer,
                            * NotifyPointerRoot, NotifyDetailNone
                            */
} XFocusChangeEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    Bool send_event;        /* true if this came from a SendEvent request */
    Display *display;       /* Display the event was read from */
    Window window;
    Atom atom;
    Time time;
    int state;              /* NewValue, Deleted */
} XPropertyEvent;

typedef struct {
    int type;
    Display *display;           /* Display the event was read from */
    XID resourceid;             /* resource id */
    unsigned long serial;       /* serial number of failed request */
    unsigned char error_code;   /* error code of failed request */
    unsigned char request_code; /* Major op-code of failed request */
    unsigned char minor_code;   /* Minor op-code of failed request */
} XErrorEvent;

typedef int (*XErrorHandler) (Display* display, XErrorEvent* event);
int XGetErrorText(Display *display, int code, char *buffer_return, int length);

typedef union {
    int type;
    XAnyEvent xany;
    XErrorEvent xerror;
    XKeyEvent xkey;
    XCreateWindowEvent xcreatewindow;
    XDestroyWindowEvent xdestroywindow;
    XFocusChangeEvent xfocus;
    XPropertyEvent xproperty;
    ...;
} XEvent;

typedef struct {
    int x, y;
    int width, height;
    int border_width;
    Window sibling;
    int stack_mode;
} XWindowChanges;

typedef struct {
    Window window;                 /* screen saver window */
    int state;                     /* ScreenSaver{Off,On,Disabled} */
    int kind;                      /* ScreenSaver{Blanked,Internal,External} */
    unsigned long til_or_since;    /* milliseconds */
    unsigned long idle;            /* milliseconds */
    unsigned long eventMask;       /* events */
} XScreenSaverInfo;


XErrorHandler XSetErrorHandler (XErrorHandler handler);

Display* XOpenDisplay(char *display_name);
int XCloseDisplay(Display *display);
int XFree(void *data);
Atom XInternAtom(Display *display, char *atom_name, Bool only_if_exists);
char* XGetAtomName(Display *display, Atom atom);

int XPending(Display *display);
int XNextEvent(Display *display, XEvent *event_return);
int XSelectInput(Display *display, Window w, long event_mask);
int XFlush(Display *display);
int XSync(Display *display, Bool discard);
Status XSendEvent(Display *display, Window w, Bool propagate, long event_mask, XEvent *event_send);

KeySym XStringToKeysym(char *string);
KeyCode XKeysymToKeycode(Display *display, KeySym keysym);

int XGrabKey(Display *display, int keycode, unsigned int modifiers,
    Window grab_window, Bool owner_events, int pointer_mode, int keyboard_mode);
int XUngrabKey(Display *display, int keycode, unsigned int modifiers, Window grab_window);

int XGetWindowProperty(Display *display, Window w, Atom property,
    long long_offset, long long_length, Bool delete, Atom req_type,
    Atom *actual_type_return, int *actual_format_return,
    unsigned long *nitems_return, unsigned long *bytes_after_return,
    unsigned char **prop_return);
int XChangeProperty(Display *display, Window w, Atom property, Atom type,
    int format, int mode, unsigned char *data, int nelements);
int XDeleteProperty(Display *display, Window w, Atom property);
int XConfigureWindow(Display *display, Window w, unsigned int value_mask,
    XWindowChanges *changes);

Status XGetGeometry(Display *display, Drawable d, Window *root_return,
    int *x_return, int *y_return, unsigned int *width_return,
    unsigned int *height_return, unsigned int *border_width_return, unsigned int *depth_return);

Status XScreenSaverQueryInfo(Display *dpy, Drawable drawable, XScreenSaverInfo *saver_info);

Status DPMSInfo (Display *display, unsigned short *power_level, unsigned char *state);
Status DPMSEnable (Display *display);
Status DPMSDisable (Display *display);

Window DefaultRootWindow(Display *display);
int ConnectionNumber(Display *display);

typedef struct {
    unsigned char   group;
    unsigned char   locked_group;
    unsigned short  base_group;
    unsigned short  latched_group;
    unsigned char   mods;
    unsigned char   base_mods;
    unsigned char   latched_mods;
    unsigned char   locked_mods;
    unsigned char   compat_state;
    unsigned char   grab_mods;
    unsigned char   compat_grab_mods;
    unsigned char   lookup_mods;
    unsigned char   compat_lookup_mods;
    unsigned short  ptr_buttons;
} XkbStateRec;

Status XkbGetState (Display *display, unsigned int device_spec, XkbStateRec *state_return);
Bool XkbLockGroup (Display *display, unsigned int device_spec, unsigned int group);

typedef struct {
    int x, y;                       /* location of window */
    int width, height;              /* width and height of window */
    int border_width;               /* border width of window */
    int depth;                      /* depth of window */
    Visual *visual;                 /* the associated visual structure */
    Window root;                    /* root of screen containing window */
    int class;                      /* InputOutput, InputOnly*/
    int bit_gravity;                /* one of the bit gravity values */
    int win_gravity;                /* one of the window gravity values */
    int backing_store;              /* NotUseful, WhenMapped, Always */
    unsigned long backing_planes;   /* planes to be preserved if possible */
    unsigned long backing_pixel;    /* value to be used when restoring planes */
    Bool save_under;                /* boolean, should bits under be saved? */
    Colormap colormap;              /* color map to be associated with window */
    Bool map_installed;             /* boolean, is color map currently installed*/
    int map_state;                  /* IsUnmapped, IsUnviewable, IsViewable */
    long all_event_masks;           /* set of events all people have interest in*/
    long your_event_mask;           /* my event mask */
    long do_not_propagate_mask;     /* set of events that should not propagate */
    Bool override_redirect;         /* boolean value for override-redirect */
    Screen *screen;                 /* back pointer to correct screen */
} XWindowAttributes;

Status XGetWindowAttributes(Display *display, Window w, XWindowAttributes *window_attributes_return);

void set_window_icon(Display *display, Window window, char *filepath);
void MagickWandGenesis(void);
void MagickWandTerminus(void);
extern "Python" int error_handler(Display* display, XErrorEvent* event);
Status XQueryTree(
    Display *display,
    Window window,
    Window *root_return,
    Window *parent_return,
    Window **children_return,
    unsigned int *nchildren_return
);
"""

LIBRARIES: List[str] = ["x11", "xscrnsaver", "xext", "MagickWand", "gdlib"]
compiler_args: Tuple[List[str], List[str]] = utils.get_compiler_args(*LIBRARIES)

ffibuilder: FFI = cffi.FFI()

ffibuilder.set_source(
    module_name="orcsome3.orcsome._xlib", source=source, extra_compile_args=compiler_args[0], libraries=compiler_args[1]
)

ffibuilder.cdef(csource=export_source, override=True)


def main(verbose: bool = False) -> None:
    ffibuilder.compile(verbose=verbose)


if __name__ == "__main__":
    main(verbose=True)



import time
from typing import Any, Callable, Optional

try:
    from ._ev import ffi, lib  # type: ignore
except ModuleNotFoundError:
    from . import ev_build

    ev_build.main()
finally:
    from ._ev import ffi, lib  # type: ignore


class Loop(object):
    def __init__(self) -> None:
        self._loop = lib.ev_loop_new(lib.EVBACKEND_SELECT)

    def destroy(self) -> None:
        lib.ev_loop_destroy(self._loop)

    def run(self, flags: int = 0) -> None:
        lib.ev_run(self._loop, flags)

    def break_(self, flags: int = lib.EVBREAK_ALL) -> None:
        lib.ev_break(self._loop, flags)


class IOWatcher(object):
    def __init__(self, callback: Callable[..., Any], file_descriptor: int, flags: int) -> None:
        self._watcher = ffi.new("ev_io*")
        self._callback = ffi.callback("io_cb", callback)
        lib.ev_io_init(self._watcher, self._callback, file_descriptor, flags)

    def start(self, loop: Loop) -> None:
        lib.ev_io_start(loop._loop, self._watcher)

    def stop(self, loop: Loop) -> None:
        lib.ev_io_stop(loop._loop, self._watcher)


class SignalWatcher(object):
    def __init__(self, callback: Callable[..., Any], signum: int) -> None:
        self._watcher = ffi.new("ev_signal*")
        self._callback = ffi.callback("signal_cb", callback)
        lib.ev_signal_init(self._watcher, self._callback, signum)

    def start(self, loop: Loop) -> None:
        lib.ev_signal_start(loop._loop, self._watcher)

    def stop(self, loop: Loop) -> None:
        lib.ev_signal_stop(loop._loop, self._watcher)


class TimerWatcher(object):
    def __init__(self, callback: Callable[..., Any], after: float, repeat: float = 0.0) -> None:
        self._after: float = after
        self._repeat: float = repeat
        self._watcher = ffi.new("ev_timer*")
        self._cb = ffi.callback("timer_cb", callback)
        self.next_stop: float = 0
        lib.ev_timer_init(self._watcher, self._cb, after, repeat)

    def start(self, loop: Loop, after: Optional[float] = None, repeat: Optional[float] = None) -> None:
        if after or repeat:
            self._after = after or self._after
            self._repeat = repeat or self._repeat
            lib.ev_timer_set(self._watcher, self._after, self._repeat)

        self.next_stop = time.time() + self._after
        lib.ev_timer_start(loop._loop, self._watcher)

    def stop(self, loop: Loop) -> None:
        lib.ev_timer_stop(loop._loop, self._watcher)

    def again(self, loop: Loop) -> None:
        self.next_stop = time.time() + self._repeat
        lib.ev_timer_again(loop._loop, self._watcher)

    def remaining(self, loop: Loop) -> float:
        return float(lib.ev_timer_remaining(loop._loop, self._watcher))

    def update_next_stop(self) -> None:
        self.next_stop = time.time() + self._repeat

    def overdue(self, timeout: float) -> bool:
        return time.time() > self.next_stop + timeout



from cffi.api import FFI as FFI
from orcsome3.orcsome import utils as utils
from typing import List, Tuple

source: str
export_source: str
LIBRARIES: List[str]
compiler_args: Tuple[List[str], List[str]]
ffibuilder: FFI

def main(verbose: bool = ...) -> None: ...



import sys
from typing import List

from setuptools import setup

import orcsome3.version


def generate_stubs() -> None:
    """
    This function re-generate the stubs for orcsome3, only meant to be used for development
    """
    try:
        from pathlib import Path

        def rmdir(directory: Path) -> None:
            """Delete all items from a directory"""
            directory = Path(directory)
            for item in directory.iterdir():
                if item.is_dir():
                    rmdir(directory=item)
                else:
                    item.unlink()
            directory.rmdir()

        def generate_files() -> Path:
            import mypy.stubgen as stubgen

            def initialize_modules(directory: Path) -> None:
                # Initializes orcsome-stubs : __init__.py
                directory.mkdir(parents=True, exist_ok=True)
                directory.joinpath("__init__.py").touch(exist_ok=True)

                # Initializes orcsome-stubs.orcsome : __init__.py
                inner_directory = directory.joinpath("orcsome")
                inner_directory.mkdir(parents=True, exist_ok=True)
                inner_directory.joinpath("__init__.py").touch(exist_ok=True)

            output_dir = Path(__file__).parent.joinpath("orcsome3-stubs")
            if output_dir.is_dir():
                rmdir(directory=output_dir)
            initialize_modules(directory=output_dir)
            args = ["--verbose", "--output", str(output_dir), str(Path(__file__).parent.joinpath("orcsome3"))]
            sys.argv.extend(args)
            # This will run "stubgen --verbose --output ./orcsome3-stubs ./orcsome3"
            stubgen.main()
            # returns the full path of ./orcsome3-stubs
            return output_dir

        def copy_files(dir: Path) -> Path:
            """
            This function will re-order the files created by function `generate_files`
            """
            import shutil

            destiny: Path = dir.joinpath("orcsome")  # ./orcsome3-stubs/orcsome
            # Copy the file ./orcsome3-stubs/orcsome3/version.pyi to ./orcsome3-stubs/version.pyi
            shutil.copy(src=dir.joinpath("orcsome3", "version.pyi"), dst=dir.joinpath("version.pyi"))
            for file_ in dir.joinpath("orcsome3", "orcsome").iterdir():
                # Copy every single file from ./orcsome3-stubs/orcsome3/orcsome to ./orcsome3-stubs/orcsome
                shutil.copy(src=file_, dst=destiny.joinpath(file_.name))
            rmdir(directory=dir.joinpath("orcsome3"))  # remove recursively ./orcsome3-stubs/orcsome3
            return destiny

        def complete_stubs(dir: Path) -> None:
            """
            This function will edit every .pyi file and replace the type Incomplete
            for the type Any, also, it will put the decorator @cached_property to the
            necessary methods accordingly to source
            """
            files_to_edit: List[Path] = []
            for file_ in dir.iterdir():
                if file_.name.endswith(".pyi"):
                    with file_.open(mode="r") as content:
                        lines = content.readlines()
                        for line in lines:
                            if (
                                "from _typeshed import Incomplete" in line
                                or "from functools import cached_property" in line
                            ):
                                files_to_edit.append(file_)
                                break

            source_files = Path(__file__).parent.joinpath("orcsome3", "orcsome")
            for file_ in files_to_edit:
                src: Path = source_files.joinpath(file_.name.replace(".pyi", ".py"))
                src_content: List[str] = []
                new_content: List[str] = []
                typing_import_src: str = ""

                with src.open(mode="r") as src_:
                    src_content = src_.readlines()
                    for line in src_content:
                        if "from typing import" in line:
                            typing_import_src = "from typing import " + ", ".join(
                                [x for x in line.replace("from typing import ", "").split(sep=", ") if x[0].isupper()]
                            )
                            break

                with file_.open(mode="r") as f_:
                    new_content = [line for line in f_.readlines() if "from _typeshed import Incomplete" not in line]
                    fix_decorator: bool = False
                    list_variables_cached: List[str] = []
                    for index, line in enumerate(iterable=new_content):
                        line_clean: str = line.translate(str.maketrans("", "", " \n\t\r"))
                        if "from typing import" in line:
                            new_content[index] = f"{typing_import_src}\n"
                        if ":Incomplete" in line_clean:
                            nombre_variable: str = line_clean.split(sep=":")[0]
                            line_fixed: bool = False
                            for src_line in src_content:
                                if f"self.{nombre_variable}:" in src_line:
                                    variable_definition: List[str] = src_line.strip().lstrip().split(sep=":")
                                    if not len(variable_definition) >= 2:
                                        continue
                                    if variable_definition[1].strip():
                                        new_content[index] = line.replace(
                                            "Incomplete", variable_definition[1].split("=")[0].strip()
                                        )
                                        line_fixed = True
                                        break
                            if not line_fixed:
                                new_content[index] = line.replace("Incomplete", "Any")
                        if "from functools import cached_property" in line:
                            fix_decorator = True
                            for index, src_line in enumerate(iterable=src_content):
                                if "@cached_property" in src_line:
                                    list_variables_cached.append(src_content[index + 1])
                        if fix_decorator and line.replace(" ...", "") in list_variables_cached:
                            new_content[index] = f"    @cached_property\n{line}"

                with file_.open(mode="w") as f_:
                    f_.writelines(new_content)

        output_dir: Path = generate_files()  # ./orcsome3-stubs
        stub_dir: Path = copy_files(dir=output_dir)
        complete_stubs(dir=stub_dir)
    except ModuleNotFoundError as e:
        print(f"EXCEPTION: {e}")


if "--generate-stubs" in sys.argv:
    sys.argv.remove("--generate-stubs")
    generate_stubs()
    exit()

setup(
    name="orcsome3",
    version=orcsome3.version.VERSION,
    author="Ahsan Pérez",
    author_email="ahsand.perez@gmail.com",
    description="Scripting extension for NETWM compliant window managers",
    long_description=open(file="README.rst").read(),
    long_description_content_type="text/x-rst",
    zip_safe=False,
    packages=["orcsome3", "orcsome3.orcsome", "orcsome3-stubs", "orcsome3-stubs.orcsome", "orcsome3.bin"],
    package_data={"orcsome3-stubs": ["*.pyi", "**/*.pyi"], "orcsome3-stubs.orcsome": ["*.pyi", "**/*.pyi"]},
    include_package_data=True,
    cffi_modules=["orcsome3/orcsome/ev_build.py:ffibuilder", "orcsome3/orcsome/xlib_build.py:ffibuilder"],
    setup_requires=["cffi>=1.0.0"],
    install_requires=["cffi>=1.0.0"],
    extras_require={"dev": ["mypy", "mypy-extensions", "typing_extensions"]},
    scripts=["orcsome3/bin/orcsome3"],
    url="https://github.com/ahsand97/orcsome3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Environment :: X11 Applications",
        "Topic :: Desktop Environment :: Window Managers",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
    ],
    python_requires=">=3.8",
)



import math
from array import array
from enum import Enum
from typing import Any, Dict, List, Optional, Union, cast

try:
    from ._xlib import ffi, lib  # type: ignore
except ModuleNotFoundError:
    from . import xlib_build

    xlib_build.main()
finally:
    from ._xlib import ffi, lib  # type: ignore

# Type Aliases
Atom = int
Window = int
Display = Any
XEvent = Any
XErrorEvent = Any
XWindowAttributes = Any
ScreenSaverInfo = Any


class MASKS(Enum):
    Mod1Mask = int(lib.Mod1Mask)
    ControlMask = int(lib.ControlMask)
    ShiftMask = int(lib.ShiftMask)
    Mod2Mask = int(lib.Mod2Mask)
    Mod4Mask = int(lib.Mod4Mask)
    LockMask = int(lib.LockMask)


class EVENTS(Enum):
    KeyPress = int(lib.KeyPress)
    KeyRelease = int(lib.KeyRelease)
    CreateNotify = int(lib.CreateNotify)
    DestroyNotify = int(lib.DestroyNotify)
    FocusIn = int(lib.FocusIn)
    FocusOut = int(lib.FocusOut)
    PropertyNotify = int(lib.PropertyNotify)
    # Unimplemented
    MapNotify = int(lib.MapNotify)
    UnmapNotify = int(lib.UnmapNotify)


class AtomCache(object):
    def __init__(self, dpy: Display):
        self.dpy: Display = dpy
        self._cache: Dict[str, Atom] = {}

    def __getitem__(self, name: str) -> Atom:
        try:
            return self._cache[name]
        except KeyError:
            pass
        finally:
            atom: Atom = lib.XInternAtom(self.dpy, str.encode(name), False)
            self._cache[name] = atom
            return atom


class XEvent_:
    class Type(Enum):
        # Types for XKeyEvent
        KeyPress = int(lib.KeyPress)
        KeyRelease = int(lib.KeyRelease)
        # Types for XCreateWindowEvent
        CreateNotify = int(lib.CreateNotify)
        # Types for XDestroyWindowEvent
        DestroyNotify = int(lib.DestroyNotify)
        # Types for XPropertyEvent
        PropertyNotify = int(lib.PropertyNotify)
        # Types for XFocusChangeEvent
        FocusIn = int(lib.FocusIn)
        FocusOut = int(lib.FocusOut)
        # Types for XCirculateEvent
        CirculateNotify = int(lib.CirculateNotify)
        # Types for XConfigureEvent
        ConfigureNotify = int(lib.ConfigureNotify)
        # Types for XGravityEvent
        GravityNotify = int(lib.GravityNotify)
        # Types for XReparentEvent
        ReparentNotify = int(lib.ReparentNotify)
        # Types for XMapEvent
        MapNotify = int(lib.MapNotify)
        # Types for XUnmapEvent
        UnmapNotify = int(lib.UnmapNotify)

    def __init__(self, xevent: XEvent, specific_event: Any) -> None:
        self._xevent: XEvent = xevent
        self.type: XEvent_.Type = self.Type(specific_event.type)


class XKeyEvent(XEvent_):
    def __init__(self, event: XEvent) -> None:
        self._xkeyevent: lib.XKeyEvent = event.xkey
        super().__init__(xevent=event, specific_event=self._xkeyevent)
        self.window: Window = self._xkeyevent.window  # "event" window it is reported relative to
        self.root: Window = self._xkeyevent.root  # root window that the event occurred on
        self.subwindow: Window = self._xkeyevent.subwindow  # child window
        self.time: float = float(self._xkeyevent.time)  # milliseconds
        self.x: int = self._xkeyevent.x  # coordinates in event window
        self.y: int = self._xkeyevent.y  # coordinates in event window
        self.x_root: int = self._xkeyevent.x_root  # coordinates relative to root
        self.y_root: int = self._xkeyevent.y_root  # coordinates relative to root
        self.state: int = self._xkeyevent.state  # key or button mask
        self.keycode: int = self._xkeyevent.keycode  # detail
        self.same_screen: bool = bool(self._xkeyevent.same_screen)  # same screen flag


class XCreateWindowEvent(XEvent_):
    def __init__(self, event: XEvent) -> None:
        self._xcreatewindowevent: lib.XCreateWindowEvent = event.xcreatewindow
        super().__init__(xevent=event, specific_event=self._xcreatewindowevent)
        self.parent: Window = self._xcreatewindowevent.parent  # parent of the window
        self.window: Window = self._xcreatewindowevent.window  # window id of window created
        self.x: int = self._xcreatewindowevent.x  # window location
        self.y: int = self._xcreatewindowevent.y  # window location
        self.width: int = self._xcreatewindowevent.width  # size of window
        self.height: int = self._xcreatewindowevent.height  # size of window
        self.border_width: int = self._xcreatewindowevent.border_width  # border width
        self.override_redirect: bool = bool(self._xcreatewindowevent.override_redirect)  # creation should be overriden


class XDestroyWindowEvent(XEvent_):
    def __init__(self, event: XEvent) -> None:
        self._xdestroywindowevent: lib.XDestroyWindowEvent = event.xdestroywindow
        super().__init__(xevent=event, specific_event=self._xdestroywindowevent)
        self.parent: Window = self._xdestroywindowevent.event
        self.window: Window = self._xdestroywindowevent.window


class XPropertyEvent(XEvent_):
    class State(Enum):
        PropertyNewValue = int(lib.PropertyNewValue)
        PropertyDelete = int(lib.PropertyDelete)

    def __init__(self, event: XEvent) -> None:
        self._xpropertyevent: lib.XPropertyEvent = event.xproperty
        super().__init__(xevent=event, specific_event=self._xpropertyevent)
        self.window: Window = self._xpropertyevent.window
        self.atom: Atom = self._xpropertyevent.atom
        self.time: float = float(self._xpropertyevent.time)
        self.state: XPropertyEvent.State = self.State(self._xpropertyevent.state)


class XFocusChangeEvent(XEvent_):
    class Mode(Enum):
        NotifyNormal = int(lib.NotifyNormal)
        NotifyWhileGrabbed = int(lib.NotifyWhileGrabbed)
        NotifyGrab = int(lib.NotifyGrab)
        NotifyUngrab = int(lib.NotifyUngrab)

    class Detail(Enum):
        NotifyAncestor = int(lib.NotifyAncestor)
        NotifyVirtual = int(lib.NotifyVirtual)
        NotifyInferior = int(lib.NotifyInferior)
        NotifyNonlinear = int(lib.NotifyNonlinear)
        NotifyNonlinearVirtual = int(lib.NotifyNonlinearVirtual)
        NotifyPointer = int(lib.NotifyPointer)
        NotifyPointerRoot = int(lib.NotifyPointerRoot)
        NotifyDetailNone = int(lib.NotifyDetailNone)

    def __init__(self, event: XEvent) -> None:
        self._xfocuschangeevent: lib.XFocusChangeEvent = event.xfocus
        super().__init__(xevent=event, specific_event=self._xfocuschangeevent)
        self.window: Window = self._xfocuschangeevent.window
        self.mode: XFocusChangeEvent.Mode = self.Mode(self._xfocuschangeevent.mode)
        self.detail: XFocusChangeEvent.Detail = self.Detail(self._xfocuschangeevent.detail)


class XErrorEvent_:
    def __init__(self, error: XErrorEvent) -> None:
        self.type: int = error.type
        self.display: Display = error.display  # Display the event was read from
        self.resourceid: float = float(error.resourceid)  # resource id
        self.serial: float = float(error.serial)  # serial number of failed request
        self.error_code: str = str(error.error_code)  # error code of failed request
        self.request_code: str = str(error.request_code)  # Major op-code of failed request
        self.minor_code: str = str(error.minor_code)  # Minor op-code of failed request

    def get_message(self, size: int = 100) -> str:
        msg = ffi.new(f"char[{size}]")
        lib.XGetErrorText(self.display, int(self.error_code), msg, size)
        pymsg: bytes = ffi.string(msg)
        return pymsg.decode()


def get_window_property(
    display: Display, window: Window, property: Atom, type: Atom = 0, split: bool = False
) -> Optional[Union[List[int], List[str]]]:
    # Default values
    # Specifies the offset in the specified property (in 32-bit quantities) where the data is to be retrieved.
    long_offset: int = 0
    long_length: int = 50  # Specifies the length in 32-bit multiples of the data to be retrieved.
    delete: bool = False  # Specifies a Boolean value that determines whether the property is deleted.

    # Params to C function XGetWindowProperty
    type_return_ = ffi.new("Atom *")
    format_return_ = ffi.new("int *")
    nitems_return_ = ffi.new("unsigned long *")
    bytes_after_ = ffi.new("unsigned long *")
    data_ = ffi.new("unsigned char **")

    result: bytes = b""

    offset: int = long_offset
    length: int = long_length
    while True:
        lib.XGetWindowProperty(
            display,
            window,
            property,
            offset,
            length,
            delete,
            type,
            type_return_,
            format_return_,
            nitems_return_,
            bytes_after_,
            data_,
        )
        format: int = format_return_[0]
        bytes_after_return: int = bytes_after_[0]
        if format == 8:  # array of usigned char
            result += ffi.buffer(data_[0], nitems_return_[0] * array("B").itemsize)
        elif format == 16:  # array of unsigned short
            result += ffi.buffer(data_[0], nitems_return_[0] * array("H").itemsize)
        elif format == 32:  # array of unsigned long
            result += ffi.buffer(data_[0], nitems_return_[0] * array("L").itemsize)
        elif not format:
            return None
        else:
            raise Exception(f"Unknown format: {format}")
        if bytes_after_return:
            lib.XFree(data_[0])
            offset = long_length
            length = math.ceil(bytes_after_return / 4 + 1)
        else:
            break

    format = format_return_[0]
    final_result: Optional[Union[List[int], List[str]]] = None
    if format == 8:
        final_result = [result.rstrip(b"\x00").decode()]
        if split:
            final_result = [x.decode() for x in result.split(b"\x00") if len(x.decode())]
    elif format == 16:
        final_result = array("H", result).tolist()
    elif format == 32:
        final_result = array("L", result).tolist()
    else:
        raise Exception(f"Unknown format: {format}")

    lib.XFree(data_[0])
    return final_result


def get_window_attributes(display: Display, window: Window) -> Optional[XWindowAttributes]:
    data_ = ffi.new("XWindowAttributes *")
    status: int = lib.XGetWindowAttributes(display, window, data_)
    if not status:
        return None
    return data_[0]


def get_screen_saver_info(display: Display, drawable: Window) -> Optional[ScreenSaverInfo]:
    data_ = ffi.new("XScreenSaverInfo *")
    status: int = lib.XScreenSaverQueryInfo(display, drawable, data_)
    if not status:
        return None
    return data_[0]


def set_window_property(
    display: Display, window: Window, property: Atom, type: Atom, format: int, values: Union[List[int], List[str]]
) -> None:
    data: Any = None
    if format == 8:  # array of usigned char
        data = cast(List[str], values)
    elif format == 16:  # array of unsigned short
        data = ffi.cast("unsigned char *", ffi.new("unsigned short[]", cast(List[int], values)))
    elif format == 32:  # array of unsigned long
        data = ffi.cast("unsigned char *", ffi.new("unsigned long[]", cast(List[int], values)))
    else:
        raise Exception(f"Unknown format {format}")

    lib.XChangeProperty(display, window, property, type, format, lib.PropModeReplace, data, len(values))


def get_kbd_group(display: Display) -> str:
    state = ffi.new("XkbStateRec *")
    lib.XkbGetState(display, lib.XkbUseCoreKbd, state)
    return str(state[0].group)


def set_kbd_group(display: Display, group: int) -> None:
    lib.XkbLockGroup(display, lib.XkbUseCoreKbd, group)
    lib.XFlush(display)


def get_atom_name(display: Display, atom: Atom) -> str:
    """
    Returns the name associated with an Atom if the Atom exists
    """
    atom_name = lib.XGetAtomName(display, atom)
    if not atom_name:
        return ""
    return bytes(ffi.string(atom_name)).decode()



from typing import Dict

KEYS: Dict[str, str] = {
    "`": "grave",
    "-": "minus",
    "=": "equal",
    "[": "bracketleft",
    "]": "bracketright",
    "\\": "backslash",
    ";": "semicolon",
    "'": "apostrophe",
    ",": "comma",
    ".": "period",
    "/": "slash",
    "bs": "BackSpace",
    "BS": "BackSpace",
    "Space": "space",
    "left": "Left",
    "right": "Right",
    "up": "Up",
    "down": "Down",
    "tab": "Tab",
    "f1": "F1",
    "f2": "F2",
    "f3": "F3",
    "f4": "F4",
    "f5": "F5",
    "f6": "F6",
    "f7": "F7",
    "f8": "F8",
    "f9": "F9",
    "f10": "F10",
    "f11": "F11",
    "f12": "F12",
    "insert": "Insert",
    "ins": "Insert",
    "delete": "Delete",
    "del": "Delete",
    "home": "Home",
    "end": "End",
    "pgup": "Prior",
    "PgUp": "Prior",
    "PageUp": "Prior",
    "pgdown": "Next",
    "PgDown": "Next",
    "PageDown": "Next",
    "menu": "Menu",
    "enter": "Return",
    "Enter": "Return",
    "return": "Return",
}



import logging
from abc import ABC, abstractmethod
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from . import ev, wrappers, xlib
from .aliases import KEYS as KEY_ALIASES

logger: logging.Logger = logging.getLogger(name=__name__)
ignore_logger: bool = False

MODIFICATORS: Dict[str, int] = {
    "Alt": xlib.MASKS.Mod1Mask.value,
    "Control": xlib.MASKS.ControlMask.value,
    "Ctrl": xlib.MASKS.ControlMask.value,
    "Shift": xlib.MASKS.ShiftMask.value,
    "Win": xlib.MASKS.Mod4Mask.value,
    "Mod": xlib.MASKS.Mod4Mask.value,
    "Hyper": xlib.MASKS.Mod4Mask.value,
    "Super": xlib.MASKS.Mod4Mask.value,
}

IGNORED_MOD_MASKS: Tuple[int, int, int, int] = (
    0,
    int(xlib.MASKS.LockMask.value),
    int(xlib.MASKS.Mod2Mask.value),
    int(xlib.MASKS.LockMask.value | xlib.MASKS.Mod2Mask.value),
)


class RestartException(Exception):
    # Exception raised by method `wm.actions.restart()`
    pass


class Actions(ABC):
    @abstractmethod
    def focus_next(self, window: Optional[wrappers.Window] = None) -> None:
        """Focus next client on current desktop.

        next/prev are defined by client creation time
        """
        pass

    @abstractmethod
    def focus_prev(self, window: Optional[wrappers.Window] = None) -> None:
        """Focus previous client on current desktop.

        next/prev are defined by client creation time
        """
        pass

    @abstractmethod
    def restart(self) -> None:
        """Restarts orcsome"""
        pass

    @abstractmethod
    def activate_window_desktop(self, window: wrappers.Window) -> Optional[bool]:
        """Activate a window's desktop

        Returns:

        * True if window is placed on different from current desktop (activates window's desktop)
        * False if window's desktop and current desktop are the same
        * None if window does not have a desktop property
        """
        pass


class WM:
    """Core orcsome instance

    Can be get in any time as::

        from orcsome3.orcsome import get_wm
        from orcsome3.orcsome.wm import WM

        wm: WM = get_wm()
    """

    def __init__(self, loop: ev.Loop) -> None:
        self._handlers: Dict[int, Callable[[xlib.XEvent], None]] = {
            xlib.EVENTS.KeyPress.value: self._handle_keypress,
            xlib.EVENTS.KeyRelease.value: self._handle_keyrelease,
            xlib.EVENTS.CreateNotify.value: self._handle_create,
            xlib.EVENTS.DestroyNotify.value: self._handle_destroy,
            xlib.EVENTS.FocusIn.value: self._handle_focus,
            xlib.EVENTS.FocusOut.value: self._handle_focus,
            xlib.EVENTS.PropertyNotify.value: self._handle_property,
        }
        # This is filled every time a new event comes by the function `_xevent_cb`
        self._native_event: Any = xlib.ffi.new("XEvent *")

        # Event associated with the callback
        self._event: Optional[
            Union[xlib.XKeyEvent, xlib.XCreateWindowEvent, xlib.XDestroyWindowEvent, xlib.XPropertyEvent]
        ] = None
        self._event_window: Optional[wrappers.Window] = None

        # Handlers
        self._key_handlers: Dict[xlib.Window, Dict[Tuple[int, int], Callable[[], None]]] = {}
        self._property_handlers: Dict[xlib.Atom, Dict[Optional[xlib.Window], List[Callable[[], None]]]] = {}
        self._create_handlers: List[Callable[[], None]] = []
        self._destroy_handlers: Dict[Optional[xlib.Window], List[Callable[[], None]]] = {}
        self._init_handlers: List[Callable[[], None]] = []
        self._deinit_handlers: List[Callable[[], None]] = []
        self._timer_handlers: List[Callable[[], None]] = []
        self._restart_handler: Optional[Callable[[], None]] = None

        # History
        self.focus_history: List[xlib.Window] = []

        # Auxiliar vars to avoid callbacks being called twice
        self._recently_destroyed_window: Optional[xlib.Window] = None
        self._recently_mapped_window: Optional[xlib.Window] = None

        self.dpy: xlib.Display = xlib.lib.XOpenDisplay(xlib.ffi.NULL)  # X11's Display
        if self.dpy == xlib.ffi.NULL:
            raise Exception("Can't open display")

        self.root: xlib.Window = xlib.lib.DefaultRootWindow(self.dpy)  # Root window
        self.atom: xlib.AtomCache = xlib.AtomCache(dpy=self.dpy)

        self._loop: ev.Loop = loop
        self._xevent_watcher: ev.IOWatcher = ev.IOWatcher(
            callback=self._xevent_cb, file_descriptor=xlib.lib.ConnectionNumber(self.dpy), flags=ev.lib.EV_READ
        )
        self._xevent_watcher.start(loop=self._loop)

        self.track_kbd_layout: bool = False
        self._startup: bool = False

        from . import actions

        self.actions: Actions = actions.Actions(window_manager=self)

    def init(self) -> None:
        # Report all events within the root window
        xlib.lib.XSelectInput(self.dpy, self.root, xlib.lib.SubstructureNotifyMask)

        for handler in self._init_handlers:
            handler()

        self._startup = True
        for window in self.get_clients():
            self._process_create_window(window=window)

        xlib.lib.XSync(self.dpy, False)
        xlib.lib.XSetErrorHandler(xlib.lib.error_handler)

        # Initializes MagickWand
        xlib.lib.MagickWandGenesis()

    def stop(self, is_exit: bool = False) -> None:
        self._key_handlers.clear()
        self._property_handlers.clear()
        self._create_handlers[:] = []
        self._destroy_handlers.clear()
        self.focus_history[:] = []

        if not is_exit:
            xlib.lib.XUngrabKey(self.dpy, xlib.lib.AnyKey, xlib.lib.AnyModifier, self.root)
            for window in self.get_clients():
                xlib.lib.XUngrabKey(self.dpy, xlib.lib.AnyKey, xlib.lib.AnyModifier, window)

        for handler in self._timer_handlers:
            getattr(handler, "stop")
        self._timer_handlers[:] = []

        for handler in self._deinit_handlers:
            try:
                handler()
            except:
                logger.exception(msg="Shutdown error")

        self._init_handlers[:] = []
        self._deinit_handlers[:] = []

        # Ends MagickWand
        xlib.lib.MagickWandTerminus()

    def create_window(self, window_id: int) -> wrappers.Window:
        window = wrappers.Window(window_id)
        window.wm = self
        return window

    def get_keycode_from_string(self, key: str) -> Optional[int]:
        keysym: int = xlib.lib.XStringToKeysym(str.encode(KEY_ALIASES.get(key, key)))
        if keysym is xlib.lib.NoSymbol:
            return None
        return int(xlib.lib.XKeysymToKeycode(self.dpy, keysym))

    def parse_keydef(self, keydef: str) -> Optional[List[Tuple[int, int]]]:
        keys: List[str] = [r.strip() for r in "".join(keydef.split()).split()]
        result: List[Tuple[int, int]] = []
        for k in keys:
            parts = k.split(sep="+")
            mod, key = parts[:-1], parts[-1]
            modmask = 0
            for m in mod:
                try:
                    modmask |= MODIFICATORS[m]
                except KeyError:
                    return None

            code = self.get_keycode_from_string(key=key)
            if not code:
                return None
            result.append((code, modmask))

        return result

    def on_key(
        self, keydef: str, window: Optional[wrappers.Window] = None
    ) -> Callable[[Callable[[], None]], Callable[[], None]]:
        """Signal decorator to define hotkey

        Signature of decorated function should be::

            function_cb() -> None:
                # ... function's body
                pass

        Key defenition is a string in format ``[mod + ... +]keysym`` where ``mod`` is
        one of modificators [Alt, Shift, Control(Ctrl), Mod(Win)] and
        ``keysym`` is a key name.

        You can define global hotkeys as follows::

            @wm.on_key(keydef='Control + a')
            def test() -> None:
                print('hotkey Control + a pressed')

            wm.on_key(keydef='Control+b')(lambda: print('hotkey Control + b pressed'))

        Or keybinded to a specific window::

            @wm.on_create(cls='URxvt')
            def bind_urxvt_keys() -> None:
                # Custom key to close only urxvt windows
                wm.on_key(window=wm.event_window, keydef='Ctrl+d')(lambda: wm.close_window(window=wm.event_window))
        """

        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                window_ = window or self.root
                code_mmask_list = self.parse_keydef(keydef=keydef)
                if not code_mmask_list or not len(code_mmask_list) == 1:
                    logger.error(msg=f"Invalid key definition [{keydef}]")
                else:
                    code, modmask = code_mmask_list[0]
                    keys: List[Tuple[int, int]] = []
                    for imask in IGNORED_MOD_MASKS:
                        mask = modmask | imask
                        xlib.lib.XGrabKey(
                            self.dpy, code, mask, window_, False, xlib.lib.GrabModeAsync, xlib.lib.GrabModeAsync
                        )
                        self._key_handlers.setdefault(window_, {})[(mask, code)] = function
                        keys.append((mask, code))

                    def remove() -> None:
                        for key in keys:
                            del self._key_handlers[window_][key]

                    setattr(function, "remove", remove)
                return function

            return inner()

        return decorator

    def on_create(self, **matchers: Any) -> Callable[[Callable[[], None]], Callable[[], None]]:
        """Signal decorator to handle window creation

        Signature of decorated function should be::

            function_cb() -> None:
                # ... function's body
                pass

        Can be used in two forms. Listen to any window creation::

           @wm.on_create()
           def debug() -> None:
               print(wm.event_window.get_wm_class())

        Or specific window::

           @wm.on_create(cls='Opera')
           def use_firefox_luke() -> None:
               wm.close_window(window=wm.event_window)
               subprocess.Popen(cmd=['firefox'])

        Also, orcsome calls on_create handlers on its startup.
        You can check ``wm.startup`` attribute to denote such event.

        See :meth:`orcsome3.orcsome.wrappers.Window.matches` for ``**matchers`` argument description.
        """

        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                return self._on_create_manage(callback=function, ignore_startup=False, **matchers)

            return inner()

        return decorator

    def on_manage(self, **matchers: Any) -> Callable[[Callable[[], None]], Callable[[], None]]:
        """Signal decorator to handle window creation (ignoring orcsome's startup)

        Signature of decorated function should be::

            function_cb() -> None:
                # ... function's body
                pass

        Can be used in two forms. Listen to any window creation::

            @wm.on_manage()
            def debug() -> None:
                print(wm.event_window.get_wm_class())

        or can be written::

            wm.on_manage()(lambda: print(wm.event_window.get_wm_class()))

        Or specific window::

            @wm.on_manage(cls='Opera')
            def use_firefox_luke():
                wm.close_window(window=wm.event_window)
                subprocess.Popen(cmd=['firefox'])

        See :meth:`orcsome3.orcsome.wrappers.Window.matches` for ``**matchers`` argument description.
        """

        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                return self._on_create_manage(callback=function, ignore_startup=True, **matchers)

            return inner()

        return decorator

    def _on_create_manage(
        self, callback: Callable[[], None], ignore_startup: bool, **matchers: Any
    ) -> Callable[[], None]:
        if matchers:
            old_function = callback

            def new_callback() -> None:
                if self.event_window.matches(**matchers):
                    old_function()

            callback = new_callback

        if ignore_startup:
            old_old_function = callback

            def new_callback() -> None:
                if not self._startup:
                    old_old_function()

            callback = new_callback

        self._create_handlers.append(callback)
        setattr(callback, "remove", lambda: self._create_handlers.remove(callback))
        return callback

    def on_destroy(
        self, window: Optional[wrappers.Window] = None
    ) -> Callable[[Callable[[], None]], Callable[[], None]]:
        """Signal decorator to handle window destroy

        Signature of decorated function should be::

            function_cb() -> None:
                # ... function's body
                pass

        It can be used to all windows::

            @wm.on_destroy()
            def cb_destroy_window() -> None:
                print(f'The window {wm.event_window} was destroyed')

        Or to a specific window::

            @wm.on_manage(name='easyeffects', cls='easyeffects')
            def on_create_easyeffects() -> None:
                @wm.on_destroy(window=wm.event_window)
                def on_destroy_easyeffects() -> None:
                    print("easyeffect's window destroyed")

        `wm.event_window` only contains the id of the recently closed window,
        trying to access any attribute of the window will result in an error
        cause the function is executed when the window is clonsing/has closed.
        """

        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                self._destroy_handlers.setdefault(window, []).append(function)

                def remove() -> None:
                    self._destroy_handlers[window].remove(function)

                setattr(function, "remove", remove)
                return function

            return inner()

        return decorator

    def on_property_change(
        self, properties: List[str], window: Optional[wrappers.Window] = None
    ) -> Callable[[Callable[[], None]], Callable[[], None]]:
        """Signal decorator to handle window property change

        Signature of decorated function should be::

            function_cb() -> None:
                # ... function's body
                pass

        One can handle any window property change::

            @wm.on_property_change(properties=['_NET_WM_STATE']) # one or multiple properties can be handled
            def window_maximized_state_change() -> None:
                window: Window = wm.event_window
                if window.maximized_vert and window.maximized_horz:
                    print('Look, ma! Window is maximized now!')

        And specific window::

            @wm.on_create()
            def switch_to_desktop() -> None:
                if not wm.startup:
                    if wm.actions.activate_window_desktop(window=wm.event_window) is None:
                        # Created window has no any attached desktop so wait for it
                        @wm.on_property_change(window=wm.event_window, properties=['_NET_WM_DESKTOP'])
                        def property_was_set() -> None:
                            wm.actions.activate_window_desktop(window=wm.event_window)
                            getattr('property_was_set', 'remove') # removes the callback

        """

        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                for prop in properties:
                    self._property_handlers.setdefault(self.atom[prop], {}).setdefault(window, []).append(function)

                def remove() -> None:
                    for prop in properties:
                        self._property_handlers[self.atom[prop]][window].remove(function)

                setattr(function, "remove", remove)
                return function

            return inner()

        return decorator

    def on_timer(
        self, timeout: float, start: bool = True, first_timeout: Optional[float] = None
    ) -> Callable[[Callable[[], None]], Callable[[], None]]:
        def decorator(function: Callable[[], None]) -> Callable[[], None]:
            @wraps(function)
            def inner() -> Callable[[], None]:
                def callback_of_timer(loop: Any, watcher: Any, events: int) -> None:
                    timer.stop(loop=self._loop) if function() else timer.update_next_stop()

                self._timer_handlers.append(function)
                timer = ev.TimerWatcher(callback=callback_of_timer, after=first_timeout or timeout, repeat=timeout)
                setattr(function, "start", lambda: timer.start(loop=self._loop))
                setattr(function, "stop", lambda: timer.stop(loop=self._loop))
                setattr(function, "again", lambda: timer.again(loop=self._loop))
                setattr(function, "remaining", lambda: timer.remaining(loop=self._loop))
                setattr(function, "overdue", lambda timeout: timer.overdue(timeout=timeout))

                if start:
                    getattr(function, "start")

                def remove() -> None:
                    try:
                        getattr(function, "stop")
                        self._timer_handlers.remove(function)
                    except:
                        pass

                setattr(function, "remove", remove)
                return function

            return inner()

        return decorator

    def get_clients(self) -> List[wrappers.Window]:
        """Return wm client list"""
        result = xlib.get_window_property(
            display=self.dpy, window=self.root, property=self.atom["_NET_CLIENT_LIST"], type=self.atom["WINDOW"] or 0
        )
        return [] if not result else [self.create_window(window_id=x) for x in cast(List[int], result)]

    def get_stacked_clients(self) -> List[wrappers.Window]:
        """Return client list in stacked order.

        Most top window will be last in list. Can be useful to determine window visibility.
        """
        result = xlib.get_window_property(
            display=self.dpy,
            window=self.root,
            property=self.atom["_NET_CLIENT_LIST_STACKING"],
            type=self.atom["WINDOW"] or 0,
        )
        return [] if not result else [self.create_window(window_id=r) for r in cast(List[int], result)]

    @property
    def current_window(self) -> Optional[wrappers.Window]:
        """Returns currently active (with input focus) window"""
        result = xlib.get_window_property(
            display=self.dpy, window=self.root, property=self.atom["_NET_ACTIVE_WINDOW"], type=self.atom["WINDOW"]
        )
        return None if not result else self.create_window(window_id=cast(List[xlib.Window], result)[0])

    @property
    def current_desktop(self) -> int:
        """Return current desktop number

        The index of the current desktop. This is always an integer between 0 and _NET_NUMBER_OF_DESKTOPS - 1
        """
        result = xlib.get_window_property(
            display=self.dpy, window=self.root, property=self.atom["_NET_CURRENT_DESKTOP"]
        )
        return cast(List[int], result)[0]

    @property
    def wm_name(self) -> Optional[str]:
        """
        Returns the actual Window Manager name or None if there's no ICCCM2.0-compliant window manager running.\n
        By the EWMH spec, a compliant window manager will set the _NET_SUPPORTING_WM_CHECK property on the root window to a window ID.\n
        If the `_NET_SUPPORTING_WM_CHECK` property exists and contains the ID of an existing window, then a ICCCM2.0-compliant window manager is running.\n
        If the property exists but does not contain the ID of an existing window, then a ICCCM2.0-compliant window manager exited without proper cleanup.\n
        If the property does not exist, then no ICCCM2.0-compliant window manager is running.
        """
        result = xlib.get_window_property(
            display=self.dpy, window=self.root, property=self.atom["_NET_SUPPORTING_WM_CHECK"], type=self.atom["WINDOW"]
        )
        if not result:
            return None
        result = xlib.get_window_property(
            display=self.dpy,
            window=cast(List[int], result)[0],
            property=self.atom["_NET_WM_NAME"],
            type=self.atom["UTF8_STRING"],
        )
        return None if not result else cast(List[str], result)[0]

    @property
    def event_window(self) -> wrappers.Window:
        """Returns the window associated with the event"""
        return cast(wrappers.Window, self._event_window)

    @property
    def event(self) -> Union[xlib.XKeyEvent, xlib.XCreateWindowEvent, xlib.XDestroyWindowEvent, xlib.XPropertyEvent]:
        """Returns the event associated"""
        return cast(
            Union[xlib.XKeyEvent, xlib.XCreateWindowEvent, xlib.XDestroyWindowEvent, xlib.XPropertyEvent], self._event
        )

    def activate_desktop(self, num: int) -> None:
        """Activate desktop ``num``"""
        if num < 0:
            return

        self._send_event(window=self.root, mtype=self.atom["_NET_CURRENT_DESKTOP"], data=[num])
        self._flush()

    def _send_event(self, window: xlib.Window, mtype: xlib.Atom, data: List[int]) -> None:
        data = (data + ([0] * (5 - len(data))))[:5]
        ev = xlib.ffi.new(
            "XClientMessageEvent *",
            {
                "type": xlib.lib.ClientMessage,
                "window": window,
                "message_type": mtype,
                "format": 32,
                "data": {"l": data},
            },
        )
        xlib.lib.XSendEvent(
            self.dpy, self.root, False, xlib.lib.SubstructureRedirectMask, xlib.ffi.cast("XEvent *", ev)
        )

    def _flush(self) -> None:
        xlib.lib.XFlush(self.dpy)

    def find_clients(self, clients: List[wrappers.Window], **matchers: Any) -> List[wrappers.Window]:
        """Return matching clients list

        :param clients: window list returned by :meth:`get_clients` or :meth:`get_stacked_clients`.
        :param **matchers: keyword arguments defined in :meth:`orcsome3.orcsome.wrappers.Window.matches`
        """
        return [r for r in clients if r.matches(**matchers)]

    def find_client(self, clients: List[wrappers.Window], **matchers: Any) -> Optional[wrappers.Window]:
        """Return first matching client

        :param clients: window list returned by :meth:`get_clients` or :meth:`get_stacked_clients`.
        :param **matchers: keyword arguments defined in :meth:`orcsome3.orcsome.wrappers.Window.matches`
        """
        result = self.find_clients(clients, **matchers)
        try:
            return result[0]
        except IndexError:
            return None

    def _process_create_window(self, window: wrappers.Window) -> None:
        xlib.lib.XSelectInput(
            self.dpy, window, xlib.lib.StructureNotifyMask | xlib.lib.PropertyChangeMask | xlib.lib.FocusChangeMask
        )
        self._event_window = window
        for handler in self._create_handlers:
            handler()

    def _handle_keypress(self, event: xlib.XEvent) -> None:
        xkeyevent: xlib.XKeyEvent = xlib.XKeyEvent(event=event)
        logger.info(msg=f"Keypress {xkeyevent.state} {xkeyevent.keycode}")
        try:
            handler = self._key_handlers[xkeyevent.window][(xkeyevent.state, xkeyevent.keycode)]
        except KeyError:
            pass
        else:
            self._event = xkeyevent
            self._event_window = self.create_window(window_id=xkeyevent.window)
            handler()

    def _handle_keyrelease(self, event: xlib.XEvent) -> None:
        keyevent: xlib.XKeyEvent = xlib.XKeyEvent(event=event)
        logger.info(msg=f"KeyRelease {keyevent.state} {keyevent.keycode}")

    def _handle_create(self, event: xlib.XEvent) -> None:
        xcreatewindowevent: xlib.XCreateWindowEvent = xlib.XCreateWindowEvent(event=event)
        self._startup = False
        window: wrappers.Window = self.create_window(window_id=xcreatewindowevent.window)
        global ignore_logger
        ignore_logger = True
        if not xlib.get_window_attributes(display=self.dpy, window=window):
            ignore_logger = False
            return
        ignore_logger = False
        self._event = xcreatewindowevent
        self._process_create_window(window=window)

    def _handle_destroy(self, event: xlib.XEvent) -> None:
        xdestroywindowevent: xlib.XDestroyWindowEvent = xlib.XDestroyWindowEvent(event=event)
        if xdestroywindowevent.window == self._recently_destroyed_window:
            return
        self._recently_destroyed_window = xdestroywindowevent.window

        handlers: List[Callable[[], None]] = []
        if None in self._destroy_handlers.keys():
            handlers.extend(self._destroy_handlers[None])
        if xdestroywindowevent.window in self._destroy_handlers.keys():
            handlers.extend(self._destroy_handlers[xdestroywindowevent.window])

        self._event = xdestroywindowevent
        self._event_window = self.create_window(window_id=xdestroywindowevent.window)
        for handler in handlers:
            handler()
        self._clean_window_data(window=xdestroywindowevent.window)

    def _handle_property(self, event: xlib.XEvent) -> None:
        xpropertyevent: xlib.XPropertyEvent = xlib.XPropertyEvent(event=event)
        atom: xlib.Atom = xpropertyevent.atom
        if xpropertyevent.state.value == xlib.lib.PropertyNewValue and atom in self._property_handlers:
            wphandlers = self._property_handlers[atom]
            self._event = xpropertyevent
            self._event_window = self.create_window(window_id=xpropertyevent.window)
            if xpropertyevent.window in wphandlers:
                for handler in wphandlers[xpropertyevent.window]:
                    handler()

            if None in wphandlers:
                for handler in wphandlers[None]:
                    handler()

    def _handle_focus(self, event: xlib.XEvent) -> None:
        xfocuschangeevent: xlib.XFocusChangeEvent = xlib.XFocusChangeEvent(event=event)
        if xfocuschangeevent.type == xlib.lib.FocusIn:
            try:
                self.focus_history.remove(xfocuschangeevent.window)
            except ValueError:
                pass

            self.focus_history.append(xfocuschangeevent.window)
            if (
                xfocuschangeevent.mode.value in (xlib.lib.NotifyNormal, xlib.lib.NotifyWhileGrabbed)
                and self.track_kbd_layout
            ):
                prop = xlib.get_window_property(
                    display=self.dpy, window=xfocuschangeevent.window, property=self.atom["_ORCSOME_KBD_GROUP"]
                )
                if prop:
                    xlib.set_kbd_group(display=self.dpy, group=int(prop[0]))
                else:
                    xlib.set_kbd_group(display=self.dpy, group=0)
        else:
            if (
                xfocuschangeevent.mode.value in (xlib.lib.NotifyNormal, xlib.lib.NotifyWhileGrabbed)
                and self.track_kbd_layout
            ):
                xlib.set_window_property(
                    display=self.dpy,
                    window=xfocuschangeevent.window,
                    property=self.atom["_ORCSOME_KBD_GROUP"],
                    type=self.atom["CARDINAL"],
                    format=32,
                    values=[xlib.get_kbd_group(display=self.dpy)],
                )

    def _xevent_cb(self, loop: Any, watcher: Any, events: int) -> None:
        event = self._native_event
        while True:
            pending_events: int = xlib.lib.XPending(self.dpy)
            if not pending_events:
                break

            while pending_events > 0:
                xlib.lib.XNextEvent(self.dpy, event)
                pending_events -= 1

                try:
                    handler = self._handlers[event.type]
                except KeyError:
                    continue

                try:
                    handler(event)
                except RestartException:
                    if self._restart_handler:
                        self._restart_handler()
                        return
                except Exception as e:
                    logger.exception(msg=e)

    def _clean_window_data(self, window: xlib.Window) -> None:
        if window in self._key_handlers:
            del self._key_handlers[window]

        if window in self._destroy_handlers:
            del self._destroy_handlers[window]

        try:
            self.focus_history.remove(window)
        except ValueError:
            pass

        for atom, whandlers in list(self._property_handlers.items()):
            if window in whandlers:
                del whandlers[window]

            if not len(self._property_handlers[atom]):
                del self._property_handlers[atom]

    def focus_window(self, window: xlib.Window) -> None:
        """Activate window"""
        self._send_event(window=window, mtype=self.atom["_NET_ACTIVE_WINDOW"], data=[2, xlib.lib.CurrentTime])
        self._flush()

    def focus_and_raise(self, window: wrappers.Window) -> None:
        """Activate window's desktop, set input focus and raise it"""
        self.actions.activate_window_desktop(window=window)
        xlib.lib.XConfigureWindow(
            self.dpy, window, xlib.lib.CWStackMode, xlib.ffi.new("XWindowChanges *", {"stack_mode": xlib.lib.Above})
        )
        self.focus_window(window=window)

    def place_window_above(self, window: xlib.Window) -> None:
        """Float up window in wm stack"""
        xlib.lib.XConfigureWindow(
            self.dpy, window, xlib.lib.CWStackMode, xlib.ffi.new("XWindowChanges *", {"stack_mode": xlib.lib.Above})
        )
        self._flush()

    def place_window_below(self, window: xlib.Window) -> None:
        """Float down window in wm stack"""
        xlib.lib.XConfigureWindow(
            self.dpy, window, xlib.lib.CWStackMode, xlib.ffi.new("XWindowChanges *", {"stack_mode": xlib.lib.Below})
        )
        self._flush()

    def _change_window_hidden_state(self, window: xlib.Window, minimize: bool) -> None:
        params = int(minimize), self.atom["_NET_WM_STATE_HIDDEN"]
        self._send_event(window=window, mtype=self.atom["_NET_WM_STATE"], data=list(params))
        self._flush()

    def minimize_window(self, window: xlib.Window) -> None:
        """Minimize window"""
        self._change_window_hidden_state(window=window, minimize=True)

    def restore_window(self, window: xlib.Window) -> None:
        """Restore window"""
        self._change_window_hidden_state(window=window, minimize=False)

    def set_window_state(
        self,
        window: xlib.Window,
        taskbar: Optional[bool] = None,
        pager: Optional[bool] = None,
        decorate: Optional[bool] = None,
        otaskbar: Optional[bool] = None,
        vmax: Optional[bool] = None,
        hmax: Optional[bool] = None,
    ) -> None:
        """
        Set window state

        `taskbar`: Indicates if the window is going to be included on the taskbar (_NET_WM_STATE_SKIP_TASKBAR)
        `pager`: Indicates if the window is going to be included on the Pager (_NET_WM_STATE_SKIP_PAGER)
        `decorate`: Indicates if the window is going to have decorations
        `otaskbar`: Indicates if the window is going to be ignored by orcsome3
        `vmax`: Indicates if the window is going to be maximized vertically (_NET_WM_STATE_MAXIMIZED_VERT)
        `hmax`: Indicates if the window is going to be maximized horizontally (_NET_WM_STATE_MAXIMIZED_HORZ)
        """
        state_atom: xlib.Atom = self.atom["_NET_WM_STATE"]
        params: Optional[List[int]] = None

        if decorate is not None:
            if self.wm_name == "Openbox":
                params = [int(not decorate), self.atom["_OB_WM_STATE_UNDECORATED"]]
                self._send_event(window=window, mtype=state_atom, data=params)
            else:
                motif_hints = xlib.get_window_property(
                    display=self.dpy, window=window, property=self.atom["_MOTIF_WM_HINTS"]
                )
                if motif_hints is not None and len(motif_hints) == 5:
                    params = cast(List[int], motif_hints)
                    params[2] = int(decorate)
                else:
                    # "0x2, 0x0, 0x0, 0x0, 0x0" to undecorate and "0x2, 0x0, 0x1, 0x0, 0x0" to redecorate
                    params = [2, 0, int(decorate), 0, 0]
                xlib.set_window_property(
                    display=self.dpy,
                    window=window,
                    property=self.atom["_MOTIF_WM_HINTS"],
                    type=self.atom["_MOTIF_WM_HINTS"],
                    format=32,
                    values=params,
                )

        if taskbar is not None:
            params = [int(not taskbar), self.atom["_NET_WM_STATE_SKIP_TASKBAR"]]
            self._send_event(window=window, mtype=state_atom, data=params)

        if vmax is not None and vmax == hmax:
            params = [int(vmax), self.atom["_NET_WM_STATE_MAXIMIZED_VERT"], self.atom["_NET_WM_STATE_MAXIMIZED_HORZ"]]
            self._send_event(window=window, mtype=state_atom, data=params)

        if otaskbar is not None:
            params = [] if otaskbar else [self.atom["_ORCSOME_SKIP_TASKBAR"]]
            xlib.set_window_property(
                display=self.dpy,
                window=window,
                property=self.atom["_ORCSOME_STATE"],
                type=self.atom["ATOM"],
                format=32,
                values=params,
            )

        if pager is not None:
            params = [int(not pager), self.atom["_NET_WM_STATE_SKIP_PAGER"]]
            self._send_event(window=window, mtype=state_atom, data=params)

        self._flush()

    def get_window_geometry(self, window: xlib.Window) -> Tuple[int, int, int, int]:
        """Get window geometry

        Returns window geometry without decorations"""
        root_ret = xlib.ffi.new("Window *")
        x = xlib.ffi.new("int *")
        y = xlib.ffi.new("int *")
        w = xlib.ffi.new("unsigned int *")
        h = xlib.ffi.new("unsigned int *")
        border_width = xlib.ffi.new("unsigned int *")
        depth = xlib.ffi.new("unsigned int *")
        xlib.lib.XGetGeometry(self.dpy, window, root_ret, x, y, w, h, border_width, depth)
        return x[0], y[0], w[0], h[0]

    def get_screen_size(self) -> Tuple[int, int]:
        """Get size of screen (root window)"""
        return self.get_window_geometry(window=self.root)[2:]

    def get_workarea(self, desktop: Optional[int] = None) -> List[int]:
        """Get workarea geometery

        :param desktop: Desktop for working area receiving. If None then current_desktop is used"""
        result = xlib.get_window_property(
            display=self.dpy, window=self.root, property=self.atom["_NET_WORKAREA"], type=self.atom["CARDINAL"]
        )
        if desktop is None:
            desktop = self.current_desktop
            if not desktop:
                return []
        return cast(List[int], result)[4 * desktop : 4 * desktop + 4]

    def moveresize_window(
        self,
        window: xlib.Window,
        x: Optional[int] = None,
        y: Optional[int] = None,
        w: Optional[int] = None,
        h: Optional[int] = None,
    ) -> None:
        """Change window geometry"""
        flags = 0
        flags |= 2 << 12
        if x is not None:
            flags |= 1 << 8
        if y is not None:
            flags |= 1 << 9
        if w is not None:
            flags |= 1 << 10
        if h is not None:
            flags |= 1 << 11
        # Workarea offsets
        o_x, o_y, _, _ = tuple(self.get_workarea())
        params = (flags, cast(int, x) + o_x, cast(int, y) + o_y, max(1, cast(int, w)), max(1, cast(int, h)))
        self._send_event(window=window, mtype=self.atom["_NET_MOVERESIZE_WINDOW"], data=list(params))
        self._flush()

    def moveresize_window2(self, window: wrappers.Window, left: int, top: int, right: int, bottom: int) -> None:
        """Change window geometry"""
        flags = 0x2F00
        # Workarea offsets
        dl, dt, dw, dh = tuple(self.get_workarea(desktop=window.desktop))
        params = (flags, left + dl, top + dt, max(1, dw - right - left), max(1, dh - bottom - top))
        self._send_event(window=window, mtype=self.atom["_NET_MOVERESIZE_WINDOW"], data=list(params))
        self._flush()

    def close_window(self, window: Optional[wrappers.Window] = None) -> None:
        """Send request to wm to close window"""
        window = window or self.current_window
        if not window:
            return
        self._send_event(window=window, mtype=self.atom["_NET_CLOSE_WINDOW"], data=[xlib.lib.CurrentTime])
        self._flush()

    def change_window_desktop(self, window: xlib.Window, desktop: int) -> None:
        """Move window to ``desktop``"""
        if desktop < 0:
            return

        self._send_event(window=window, mtype=self.atom["_NET_WM_DESKTOP"], data=[desktop])
        self._flush()

    def on_init(self, func: Callable[[], None]) -> Callable[[], None]:
        """
        Adds a function to the list of init handlers, every function on the list gets
        executed whenever orcsome3 is starting
        """
        self._init_handlers.append(func)
        return func

    def on_deinit(self, func: Callable[[], None]) -> Callable[[], None]:
        """
        Adds a function to the list of de-init handlers, every function on the list gets
        executed whenever orcsome3 is stopping
        """
        self._deinit_handlers.append(func)
        return func

    def get_screen_saver_info(self) -> Optional[wrappers.XScreenSaverInfo]:
        """
        This is a wrapper around `XScreenSaverQueryInfo`, returns information about the current state of the screen saver or `None` if there's no screensaver
        active.
        """
        xscreensaverinfo = xlib.get_screen_saver_info(display=self.dpy, drawable=self.root)
        if not xscreensaverinfo:
            return None
        return wrappers.XScreenSaverInfo(screensaverinfo=xscreensaverinfo)

    def reset_dpms(self) -> None:
        """
        Resets Display Power Management Signaling (DPMS)
        """
        power = xlib.ffi.new("unsigned short *")
        state = xlib.ffi.new("unsigned char *")
        if xlib.lib.DPMSInfo(self.dpy, power, state):
            if state[0]:
                xlib.lib.DPMSDisable(self.dpy)
                xlib.lib.DPMSEnable(self.dpy)

    def get_atom_name(self, atom: xlib.Atom) -> str:
        """
        Return the name associated with an atom
        """
        return xlib.get_atom_name(display=self.dpy, atom=atom)

    def _set_window_icon(self, window: xlib.Window, icon: str) -> None:
        xlib.lib.set_window_icon(self.dpy, window, xlib.ffi.new("char[]", icon.encode()))  # char[] is char*

    def _get_window_tree(
        self, window: wrappers.Window
    ) -> Optional[Tuple[wrappers.Window, wrappers.Window, List[wrappers.Window], int]]:
        root_window_ = xlib.ffi.new("Window *")
        parent_window_ = xlib.ffi.new("Window *")
        children_windows_ = xlib.ffi.new("Window **")
        number_of_children_ = xlib.ffi.new("unsigned int *")

        result: int = xlib.lib.XQueryTree(
            self.dpy, window, root_window_, parent_window_, children_windows_, number_of_children_
        )
        if not result:
            return None
        children_windows: List[wrappers.Window] = []
        for i in range(0, number_of_children_[0], 1):
            window_: wrappers.Window = self.create_window(window_id=children_windows_[0][i])
            children_windows.append(window_)
        return (
            self.create_window(window_id=root_window_[0]),
            self.create_window(window_id=parent_window_[0]),
            children_windows,
            number_of_children_[0],
        )


class ImmediateWM(WM):
    """
    ImmediateWM which allows to play with wm from repl.
    """

    def __init__(self) -> None:
        self.dpy: xlib.Display = xlib.lib.XOpenDisplay(xlib.ffi.NULL)
        if self.dpy == xlib.ffi.NULL:
            raise Exception("Can't open display")

        self.root: xlib.Window = xlib.lib.DefaultRootWindow(self.dpy)
        self.atom: xlib.AtomCache = xlib.AtomCache(dpy=self.dpy)


@xlib.ffi.def_extern()  # type: ignore
def error_handler(_display: xlib.Display, error: xlib.XErrorEvent) -> int:
    err: xlib.XErrorEvent_ = xlib.XErrorEvent_(error=error)
    logger.error(
        msg=f"{err.get_message()} ({err.request_code}:{err.minor_code}) ({'0x%0.2X' % int(err.resourceid)}:{int(err.resourceid)})"
    ) if not ignore_logger else None
    return 0



import cffi
from cffi.api import FFI

source: str = """
#include <ev.h>
"""

export_source: str = """
#define EVBACKEND_SELECT ...
#define EV_READ ...
#define EV_WRITE ...
#define EVBREAK_ALL ...

typedef ... ev_loop;

struct ev_loop *ev_loop_new (unsigned int flags);
void ev_loop_destroy (struct ev_loop*);
void ev_break (struct ev_loop*, int);
int ev_run (struct ev_loop*, int);

typedef struct { ...; } ev_io;
typedef void (*io_cb) (struct ev_loop*, ev_io*, int);
void ev_io_init(ev_io*, io_cb, int, int);
void ev_io_start(struct ev_loop*, ev_io*);
void ev_io_stop(struct ev_loop*, ev_io*);

typedef struct { ...; } ev_signal;
typedef void (*signal_cb) (struct ev_loop*, ev_signal*, int);
void ev_signal_init(ev_signal*, signal_cb, int);
void ev_signal_start(struct ev_loop*, ev_signal*);
void ev_signal_stop(struct ev_loop*, ev_signal*);

typedef double ev_tstamp;
typedef struct { ...; } ev_timer;
typedef void (*timer_cb) (struct ev_loop*, ev_timer*, int);
void ev_timer_init(ev_timer*, timer_cb, ev_tstamp, ev_tstamp);
void ev_timer_set(ev_timer*, ev_tstamp, ev_tstamp);
void ev_timer_start(struct ev_loop*, ev_timer*);
void ev_timer_again(struct ev_loop*, ev_timer*);
void ev_timer_stop(struct ev_loop*, ev_timer*);
ev_tstamp ev_timer_remaining(struct ev_loop*, ev_timer*);
"""

ffibuilder: FFI = cffi.FFI()

ffibuilder.set_source(module_name="orcsome3.orcsome._ev", source=source, libraries=["ev"])

ffibuilder.cdef(csource=export_source, override=True)


def main(verbose: bool = False) -> None:
    ffibuilder.compile(verbose=verbose)


if __name__ == "__main__":
    main(verbose=True)



from . import wm as wm, wrappers as wrappers
from typing import Optional

class Actions(wm.Actions):
    def __init__(self, window_manager: wm.WM) -> None: ...
    def focus_next(self, window: Optional[wrappers.Window] = ...) -> None: ...
    def focus_prev(self, window: Optional[wrappers.Window] = ...) -> None: ...
    def restart(self) -> None: ...
    def activate_window_desktop(self, window: wrappers.Window) -> Optional[bool]: ...
